// MgaLibOps.cpp : Implementation of Mga library operations
#include "stdafx.h"
#include "MgaFCO.h"
#include "MgaSet.h"
#include "MgaReference.h"
#include "MgaConnection.h"
#include "MgaLibOps.h" // by ZolMol


HRESULT FCO::get_LibraryName(BSTR *libname) {
	COMTRY {
		CHECK_OUTSTRPAR(libname);
		CheckRead();
		if(self[ATTRID_PERMISSIONS] & LIBROOT_FLAG) {
			CComBSTR nn = self[ATTRID_NAME];
			*libname = nn.Detach();
		}
		else *libname = NULL;
	} COMCATCH(;);
}


void setlibpermall(CoreObj &c, long newmask) {
	ASSERT(newmask == LIBRARY_FLAG || newmask == 0);
	ITERATE_THROUGH(c[ATTRID_PARENT+ATTRID_COLLECTION]) {
		ITER[ATTRID_PERMISSIONS] = (ITER[ATTRID_PERMISSIONS] & INSTANCE_FLAG) | newmask;
		if(ITER.IsContainer()) setlibpermall(ITER, newmask);
	}
}	

HRESULT FCO::put_LibraryName(BSTR libname) {
	COMTRY_IN_TRANSACTION {
		CHECK_INSTRPAR(libname);
		CheckWrite();
		if(CComBSTR(libname)) {
			self[ATTRID_PERMISSIONS] = LIBROOT_FLAG;
			self[ATTRID_NAME] = libname;
			setlibpermall(self, LIBRARY_FLAG);
			SelfMark(OBJEVENT_PROPERTIES);
		}
		else {
			self[ATTRID_PERMISSIONS] = 0;
			setlibpermall(self, 0);
		}
	} COMCATCH_IN_TRANSACTION(;);
}

/* *************************** Pointer Fixup ************************************** */
/*
by ZolMol
class PointerFixup
{
	coreobjpairhash identity_map;

	typedef struct
	{
		CoreObj target;
		CoreObj created;
		attrid_type attrid;
	} resolve_entry;

	std::vector<resolve_entry> resolve_entries;

public:
	inline void identify(const CoreObj &original, const CoreObj &created)
	{
		ASSERT( original != NULL && created != NULL );

		identity_map.insert(coreobjpairhash::value_type(original, created));
	}

	void resolve(const CoreObj &original, const CoreObj &created,
		attrid_type attrid)
	{
		if( LINKREF_ATTR(attrid) )
			return;

		CoreObj target = original[attrid]; 
		if( target == NULL )
			return;

		coreobjpairhash::iterator i = identity_map.find(target);
		if( i != identity_map.end() )
			created[attrid] = i->second;	// already created
		else								// resolve it later
		{
			resolve_entry entry;
			entry.target = target;
			entry.created = created;
			entry.attrid = attrid;
			resolve_entries.push_back(entry);
		}
	}

	inline void clear()
	{
		resolve_entries.clear();
		identity_map.clear();
	}

	void fixPointers()
	{
		std::vector<resolve_entry>::const_iterator i = resolve_entries.begin();
		while( i != resolve_entries.end() )
		{
			CoreObj& new_target = identity_map[i->target];
			ASSERT( new_target != NULL );

			(i->created)[i->attrid] = new_target;

			++i;
		}
		clear();
	}
};
*/
/* *************************** Attach ********************************************* */

void ObjTreeCopyFromExt(CMgaProject *mgaproject, const CoreObj &orig, CoreObj &nobj, PointerFixup &fixup) {
	metaid_type s;
	COMTHROW(mgaproject->dataproject->CreateObject(s = orig.GetMetaID(), &nobj.ComPtr()));
	
	if( s != DTID_REGNODE )		// speedup, no references to regnodes
		fixup.identify(orig, nobj);

	if(s >= DTID_MODEL && s <= DTID_FOLDER) {
		setcheck(mgaproject, nobj, CHK_NEW);
		CoreObjMark(nobj, OBJEVENT_CREATED);
	}

	CComPtr<ICoreAttributes> atts;
	COMTHROW(orig->get_Attributes(&atts));
	MGACOLL_ITERATE(ICoreAttribute, atts) {
			attrid_type ai;
			CComPtr<ICoreMetaAttribute> mattr;
			COMTHROW(MGACOLL_ITER->get_MetaAttribute(&mattr));
			COMTHROW(mattr->get_AttrID(&ai));
			if(ai < ATTRID_COLLECTION) {
				valtype_type mvt;
				COMTHROW(mattr->get_ValueType(&mvt));
				if(mvt != VALTYPE_POINTER) {
					if(ai == ATTRID_PERMISSIONS) {
						nobj[ai] = orig[ai] & INSTANCE_FLAG | LIBRARY_FLAG;
					}
					else nobj[ai] = static_cast<CComVariant>(orig[ai]);
				}
				else {
					fixup.resolve(orig, nobj, ai);
				}
			}
			else {
				ai -= ATTRID_COLLECTION;
				if(LINKREF_ATTR(ai)) {
					CoreObjs collmembers = orig[ai + ATTRID_COLLECTION];
					ITERATE_THROUGH(collmembers) {
						CoreObj nchild;
						ObjTreeCopyFromExt(mgaproject, ITER, nchild, fixup);
						nchild[ai] = nobj;
					}
				}
			}

	} MGACOLL_ITERATE_END;
}

void CreateLibraryImage(CMgaProject *mgaproject, CComBSTR &connstr, CoreObj &libimgroot) {
	CComPtr<IMgaProject> p;
	PointerFixup fixup;
	try {
		COMTHROW(p.CoCreateInstance(CLSID_MgaProject));
		
		CComBSTR paradigmname;
		CComVariant paradigmGUID;
		COMTHROW(mgaproject->get_MetaName(&paradigmname));
		COMTHROW(mgaproject->get_MetaGUID(&paradigmGUID));

		COMTHROW(p->OpenEx(connstr, paradigmname, paradigmGUID));
		COMTHROW(p->BeginTransaction(NULL, TRANSACTION_READ_ONLY));
		{
			CComPtr<IMgaFolder> libroot;
			COMTHROW(p->get_RootFolder(&libroot));
	
			ObjTreeCopyFromExt(mgaproject, CoreObj(libroot), libimgroot, fixup);
		}
		fixup.fixPointers();

		libimgroot[ATTRID_NAME] = connstr;
		libimgroot[ATTRID_PERMISSIONS] = LIBROOT_FLAG;

		COMTHROW(p->AbortTransaction());
		COMTHROW(p->Close());
	} catch(...) { 
		fixup.clear();
		if(p) {
			long st;
			p->get_ProjectStatus(&st);
			if(st & 8) p->AbortTransaction();
			if(st & 1) p->Close();
		};
		throw;
	}
}

HRESULT FCO::AttachLibrary(BSTR libname, IMgaFolder **f, long relid) {
	COMTRY_IN_TRANSACTION {
		CHECK_OUTPTRPARVALIDNULL(f);
			
		CoreObj libimgroot;
		CreateLibraryImage(mgaproject, CComBSTR(libname), libimgroot);

		assignnewchild(libimgroot);

		docheck(mgaproject);

		if(f) ObjForCore(libimgroot)->getinterface(f);

		ObjTreeNotify(mgaproject, libimgroot);
		SelfMark(OBJEVENT_NEWCHILD);
	} COMCATCH_IN_TRANSACTION(;);
}

/* *************************** Refresh ********************************************* */

void steal(CoreObj &o, CoreObj &n, attrid_type ai) {
	ITERATE_THROUGH(o[ai + ATTRID_COLLECTION]) {
		ITER[ai] = n;
	}
}

void redo_derivs(CMgaProject *mgaproject, CoreObj &oldnode, CoreObj &newnode, bool deriv_non_primary) {
	bool libnode = (newnode[ATTRID_PERMISSIONS] & LIBRARY_FLAG) ? true : false;
	if(long(oldnode[ATTRID_META]) != long(newnode[ATTRID_META])) {
		COMTHROW(E_MGA_LIB_DIFF);
	}
	if(oldnode.IsFCO()) {
		CoreObjs ders = oldnode[ATTRID_DERIVED+ATTRID_COLLECTION];
		ITERATE_THROUGH(ders) {
			long perm = ITER[ATTRID_PERMISSIONS];
			if(perm & LIBRARY_FLAG) continue;

			if(!deriv_non_primary && ITER[ATTRID_RELID] >= RELIDSPACE) continue;  // not primary derived
// 
			CoreObj newder;
			coreobjpairhash crealist;
			ObjTreeDerive(mgaproject, newnode, newder, crealist, perm & INSTANCE_FLAG);  
			newder[ATTRID_ROLEMETA] = ITER[ATTRID_ROLEMETA];  // maybe 0
			newder[ATTRID_PARENT] = ITER[ATTRID_PARENT];
			newder[ATTRID_RELID] = ITER[ATTRID_RELID];
							
			ObjTreeReconnect(newder, crealist); // Reroute references
			redo_derivs(mgaproject, ITER, newder, true);  // RECURSION!
		}
	}
	if(oldnode.IsContainer()) {
		CoreObjs oldchildren = oldnode[ATTRID_PARENT+ATTRID_COLLECTION];
		Sort<long>(oldchildren, ATTRID_RELID);
		CoreObjs newchildren = newnode[ATTRID_PARENT+ATTRID_COLLECTION];
		Sort<long>(newchildren, ATTRID_RELID);
		std::vector<CoreObj> newvect;
		GetAll2<ICoreObjects>(newchildren, newvect);
		std::vector<CoreObj>::iterator iter = newvect.begin(), iter_end = newvect.end();
		if(iter != iter_end) {
			long newrelid = (*iter)[ATTRID_RELID];
			ITERATE_THROUGH(oldchildren) {
				long oldrelid = ITER[ATTRID_RELID];
				while(newrelid < oldrelid && iter != iter_end )
				{
					if( ++iter != iter_end )
						newrelid = (*iter)[ATTRID_RELID];
				}
				// no refresh for objects with 0 relID
				if(oldrelid != 0 && oldrelid == newrelid) {
					redo_derivs(mgaproject, ITER, *iter, false);
				}
				else if(oldrelid < RELIDSPACE && !libnode) {
					ITER[ATTRID_PARENT] = newnode;
				}
			}
		}
	}
	
	// PETER : reroute references from the model
	if (libnode && oldnode.IsFCO()) {
		// Sets ? COnnections ? Other relations ?
		ITERATE_THROUGH(oldnode[ATTRID_REFERENCE + ATTRID_COLLECTION]) {
			if (ITER[ATTRID_PERMISSIONS] & LIBRARY_FLAG) continue;
			ITER[ATTRID_REFERENCE] = newnode;
		}
	}
	// PETER END

	if (libnode) 
		return;

	ASSERT(newnode[ATTRID_RELID] == oldnode[ATTRID_RELID]);
	ASSERT(newnode[ATTRID_META] == oldnode[ATTRID_META]);

	newnode[ATTRID_NAME] = oldnode[ATTRID_NAME];
	if(oldnode.IsContainer()) {
		newnode[ATTRID_LASTRELID] = oldnode[ATTRID_LASTRELID];
	}
	steal(oldnode, newnode, ATTRID_REGNOWNER);
	steal(oldnode, newnode, ATTRID_CONSTROWNER);

	if(oldnode.IsFCO()) {
		ASSERT(newnode[ATTRID_ROLEMETA] == oldnode[ATTRID_ROLEMETA]);
		steal(oldnode, newnode, ATTRID_ATTRPARENT);
		steal(oldnode, newnode, ATTRID_XREF);
		steal(oldnode, newnode, ATTRID_REFERENCE);
		switch(oldnode.GetMetaID()) {
		case DTID_REFERENCE:
			steal(oldnode, newnode, ATTRID_SEGREF);
			if(!CoreObj(oldnode[ATTRID_MASTEROBJ])) {
				std::vector<CoreObj> peer(1);
				peer[0] = oldnode[ATTRID_REFERENCE];   // set the old val, it will
													   // be updated through xref
				newnode[ATTRID_MASTEROBJ] = NULLCOREOBJ;
				putreftask(false).DoWithDeriveds(newnode, &peer);
			}
			break;
		case DTID_CONNECTION:
			{
				ITERATE_THROUGH(oldnode[ATTRID_CONNROLE + ATTRID_COLLECTION]) {
					if(!CoreObj(ITER[ATTRID_MASTEROBJ])) {
						CoreObj rr;
						if(ObjForCore(newnode)->findroleobj(CComBSTR(ITER[ATTRID_NAME]), rr)) {
							RemoveConnPTask().DoWithDeriveds(rr);
						}
// todo: propagate
						ITER[ATTRID_CONNROLE] = newnode;
					}
				}
			}
			break;
		case DTID_SET:
			{
				ITERATE_THROUGH(oldnode[ATTRID_SETMEMBER + ATTRID_COLLECTION]) {
					if(!CoreObj(ITER[ATTRID_MASTEROBJ])) {
// todo: delete duplicate setnodes
// todo: propagate
						ITER[ATTRID_SETMEMBER] = newnode;
					}
				}
			}
			break;
		}
	}
}


HRESULT FCO::RefreshLibrary(BSTR libname) {
	COMTRY_IN_TRANSACTION {
		CheckWrite();
		if(!(self[ATTRID_PERMISSIONS] & LIBROOT_FLAG)) {
			COMTHROW(E_MGA_OP_REFUSED);
		}

		CComBSTR ln(libname);
		if(!ln) COMTHROW(get_LibraryName(&ln));
		CoreObj libimgroot;
		CreateLibraryImage(mgaproject, ln, libimgroot);
 
		libimgroot[ATTRID_PARENT] = self[ATTRID_PARENT];
		libimgroot[ATTRID_RELID] = self[ATTRID_RELID];

		redo_derivs(mgaproject, self, libimgroot, false);
		inDeleteObject();
		docheck(mgaproject);
	} COMCATCH_IN_TRANSACTION(;);
}

