#include "stdafx.h"
#include "Parser.h"
#include "MgaParser.h"
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <stdio.h>

#include "../Common/CommonCollection.h"
#include "../Common/CommonMgaTrukk.h"
//#include <list>//slist
#include "helper.h"

void CMgaParser::findObjOnAbsPath( CComObjPtr<IMgaProject> p_project, const std::string& p_absPath , CComObjPtr<IMgaObject>& p_obj, const std::string& text)
{
	COMTHROW( p_project->get_ObjectByPath( PutInBstr( p_absPath), PutOut( p_obj)) );
	if( !p_obj) // if nonunique is allowed
	{
		COMTHROW( p_project->get_NthObjectByPath( 0, PutInBstr( p_absPath), PutOut( p_obj)) );

		if( p_obj)
		{
			// name ambiguity exists for sure
			//<!> warning about non uniqueness
			if( m_GME) 
			{
				CComBSTR bstr, id, nm;
				COMTHROW( p_obj->get_ID( &id));
				COMTHROW( p_obj->get_Name( &nm));
				bstr.Append("[Smartcopy Parser] Name ambiguity, selected: <A HREF=\"mga:");
				bstr.AppendBSTR( id);
				bstr.Append("\">");
				if( nm.Length()!= 0)
					bstr.AppendBSTR( nm);
				else
					bstr.Append("emptyname");
				bstr.Append("</A>");
				
				if( !text.empty())
				{
					bstr.Append(" as ");
					bstr.Append( text.c_str());
				}

				bstr.Append(". Search path used: ");
				bstr.Append( makeViewable(p_absPath).c_str());

				COMTHROW( m_GME->ConsoleMessage(bstr, MSG_WARNING));
			}
		}
	}
}

void CMgaParser::findObjOnRelPath( CComObjPtr<IMgaObject> obj_rel_to, const std::string& relpath , CComObjPtr<IMgaObject>& obj, const std::string& text)
{
	ASSERT( !relpath.empty());
	ASSERT( obj_rel_to);

	CComObjPtr<IMgaFCO> cur_fco;
	CComObjPtr<IMgaFolder> cur_fld;

	obj_rel_to.QueryInterface( cur_fco);
	obj_rel_to.QueryInterface( cur_fld);

	bool isfco = (cur_fco != NULL);

	std::string m = relpath;
	
	while( m.substr(0,4) == "/@..")
	{
		CComObjPtr<IMgaModel> mod;
		if( isfco) COMTHROW( cur_fco->get_ParentModel( PutOut( mod)) );

		if ( mod)	// has a valid model parent
		{
			cur_fco = mod;
		}
		else		// must have folder parent
		{
			CComObjPtr<IMgaFolder> fold;

			if( isfco)  COMTHROW( cur_fco->get_ParentFolder( PutOut( fold)) );
			else		COMTHROW( cur_fld->get_ParentFolder( PutOut( fold)) );
			
			if ( fold)
			{
				isfco = false;
				cur_fld = fold;
			}
			// else rootfolder reached: wrong path information
		}
		
		m = m.substr(4);
	}
	
	if( m.empty())
	{
		if (isfco)	obj = cur_fco;
		else 		obj = cur_fld;
	}
	else if( isfco)
		COMTHROW( cur_fco->get_ObjectByPath( PutInBstr( m), PutOut( obj)) );
	else
		COMTHROW( cur_fld->get_ObjectByPath( PutInBstr( m), PutOut( obj)) );

	if( !obj) // not found , name ambiguity may exist, take the 1st
	{
		if( isfco)
			COMTHROW( cur_fco->get_NthObjectByPath( 0, PutInBstr( m), PutOut( obj)) );
		else
			COMTHROW( cur_fld->get_NthObjectByPath( 0, PutInBstr( m), PutOut( obj)) );

		if( obj) // name ambiguity exists for sure
		{
			if( m_GME) 
			{
				CComBSTR bstr, id, nm;
				COMTHROW( obj->get_ID( &id));
				COMTHROW( obj->get_Name( &nm));
				bstr.Append("[Smartcopy Parser] Name ambiguity, selected: <A HREF=\"mga:");
				bstr.AppendBSTR( id);
				bstr.Append("\">");
				if( nm.Length()!= 0)
					bstr.AppendBSTR( nm);
				else
					bstr.Append("emptyname");
				bstr.Append("</A>"); 

				if( !text.empty())
				{
					bstr.Append(" as ");
					bstr.Append( text.c_str());
				}

				bstr.Append(". Search path used: ");
				bstr.Append( makeViewable(relpath).c_str());

				COMTHROW( m_GME->ConsoleMessage(bstr, MSG_WARNING));
			}
		}
	}
}

bool CMgaParser::findPlaceForElem( 
							     const attributes_type &attributes
							   , deriv_type& deriv
							   , CComObjPtr<IMgaObject>& place
							)
{
	const std::string *nm = GetByNameX(attributes, "closurepath");
	if( nm != NULL)
	{
		if( nm->compare("") == 0)
		{
			place = m_target;
			return true;
		}
		else
		{
			findObjOnRelPath( m_target, *nm, place, "place");
			if( place)
				return true;
			else
				ASSERT(0);
		}
	}
	return false;
}

//<!> check if findObjOnRelPath is given "" string

bool CMgaParser::findPlaceForElem( 
							     const attributes_type &attributes
							   , deriv_type& deriv
							   , CComObjPtr<IMgaFolder>& place_folder
							   , CComObjPtr<IMgaModel>& place_model
							)
{
	
	CComObjPtr<IMgaObject> place;
	if( findPlaceForElem( attributes, deriv, place))
	{
		place.QueryInterface( place_folder);
		place.QueryInterface( place_model);
		return (place_folder || place_model);
	}
	else
	{
		place = m_target;//<!> let's try this
		
		CComBSTR id, nm, bstr("[Smartcopy Parser] Correct place not found for object: ");
		bstr.Append( makeNameViewable( GetByName( attributes, "closurename")).c_str());
		bstr.Append(". Search path used: ");
		bstr.Append( makeViewable( GetByName(attributes, "closurepath")).c_str());
		bstr.Append(". Trying to insert into the target object: <A HREF=\"mga:");
		COMTHROW( place->get_ID( &id));
		COMTHROW( place->get_Name( &nm));
		bstr.AppendBSTR( id);
		bstr.Append("\">");
		if( nm.Length() != 0)
			bstr.AppendBSTR( nm);
		else
			bstr.Append("emptyname");
		bstr.Append("</A>.");

		if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));

		place.QueryInterface( place_folder);
		place.QueryInterface( place_model);
		return (place_folder || place_model);
	}
	return false;
}

/*static*/ CComPtr<IGMEOLEApp> CMgaParser::get_GME(CComObjPtr<IMgaProject> project)
{
	CComPtr<IGMEOLEApp> gme;
	if ( (project != NULL)) {		
		CComBSTR bstrName("GME.Application");
		CComPtr<IMgaClient> pClient;
		HRESULT hr = project->GetClientByName(bstrName, &pClient);
		if (SUCCEEDED(hr) && pClient) {
			CComPtr<IDispatch> pDispatch;
			hr = pClient->get_OLEServer(&pDispatch);
			if (SUCCEEDED(hr) && pDispatch) {
				hr = pDispatch.QueryInterface(&gme);
				if (FAILED(hr)) {
					gme = NULL;
				}
			}
		}
	}
	return gme;
} 

/*static*/ void CMgaParser::clear_GME( CComPtr<IGMEOLEApp>& p_GME)
{
	p_GME = CComPtr<IGMEOLEApp>();
}

void CMgaParser::popInfo()
{
	if( m_parsed1.empty() || m_parsed2.empty()) return; // empty stack of informations

	CComObjPtr<IMgaObject> obj = m_parsed1[0]; m_parsed1.pop_back();
	CComObjPtr<IMgaObject> fco1 = m_parsed2[0]; m_parsed2.pop_back();
	CComObjPtr<IMgaObject> fco2 = m_parsed3[0]; m_parsed3.pop_back();

	writeInfo( obj, fco1, fco2);
}

void CMgaParser::pushInfo( const CComObjPtr<IMgaFolder>& place_fld, const CComObjPtr<IMgaModel>& place_mdl, const CComObjPtr<IMgaFCO>& fco1, const CComObjPtr<IMgaFCO>& fco2)
{
	CComObjPtr<IMgaObject> obj;
	if( place_fld)
		obj = place_fld;
	else if( place_mdl)
		obj = place_mdl;
	else ASSERT(0);

	//m_parsed.push_back( pair< CComObjPtr<IMgaObject>, CComObjPtr<IMgaObject>( obj, fco1) );
	m_parsed1.push_back( obj);
	m_parsed2.push_back( CComObjPtr<IMgaObject>( fco1));
	m_parsed3.push_back( CComObjPtr<IMgaObject>( fco2));
}

void CMgaParser::writeInfo( const CComObjPtr<IMgaObject>& obj, const CComObjPtr<IMgaObject>& fco1, const CComObjPtr<IMgaObject>& fco2, bool normal_msg)
{
	if( obj && fco1 )
	{
		CComBSTR bstr, id2, nm2, id1, nm1, id0, nm0;
		COMTHROW( obj->get_ID( &id0));
		COMTHROW( obj->get_Name( &nm0));
		COMTHROW( fco1->get_ID( &id1));
		COMTHROW( fco1->get_Name( &nm1));
		if( fco2) COMTHROW( fco2->get_ID( &id2));
		if( fco2) COMTHROW( fco2->get_Name( &nm2));

		bstr.Append("[Smartcopy Parser] <A HREF=\"mga:");
		bstr.AppendBSTR( id1);
		bstr.Append("\">");

		if( nm1.Length() != 0)	bstr.AppendBSTR( nm1);
		else					bstr.Append("emptyname");

		bstr.Append( "</A>");

		if( fco2) 
		{
			bstr.Append( " derived from <A HREF=\"mga:");
			bstr.AppendBSTR( id2);
			bstr.Append("\">");

			if( nm2.Length() != 0)	bstr.AppendBSTR( nm2);
			else					bstr.Append("emptyname");

			bstr.Append("</A>");
		}

		if( !normal_msg)
			bstr.Append(" could not be");
		bstr.Append( " inserted into <A HREF=\"mga:");
		bstr.AppendBSTR( id0);
		bstr.Append( "\">");

		if( nm0.Length() != 0)	bstr.AppendBSTR( nm0);
		else					bstr.Append("emptyname");

		bstr.Append( "</A>.");
		if( m_GME) 
		{
			if( normal_msg) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_INFO));
			else			COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
		}
	}
}