#include "stdafx.h"
#include "Parser.h"
#include "MgaParser.h"
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <stdio.h>

#include "../Common/CommonCollection.h"
#include "../Common/CommonMgaTrukk.h"
#include <list>//slist
#include "helper.h"

// --------------------------- CMgaParser - copy closure dumped data parsing

STDMETHODIMP CMgaParser::ParseClos4(IMgaObject *here, BSTR filename)
{
	CHECK_IN(here);

	try
	{
		CloseAll();

		COMTHROW( progress.CoCreateInstance(L"Mga.MgaProgressDlg") );
		COMTHROW( progress->SetTitle(PutInBstr("Importing SmartCopied XML data...")) );
		COMTHROW( progress->StartProgressDialog(NULL) );

		CComObjPtr<IMgaProject> p;
		COMTHROW( here->get_Project(PutOut(p)) );
		ASSERT( p != NULL );
		COMTHROW( p->get_Preferences(&project_prefs_orig) );
		manual_relid_mode = project_prefs_orig & MGAPREF_MANUAL_RELIDS ? true : false;

		m_GME = get_GME( p);//by zolmol

		COMTHROW( p->CreateTerritory(NULL, PutOut(territory), NULL) );
		COMTHROW( p->BeginTransaction(territory, TRANSACTION_NON_NESTED) );

		CComObjPtr<IMgaObject> target;
		COMTHROW(territory->OpenObj(here, PutOut(target)));

		//by zolmol
		m_target = target;
		m_resolveDerFuncPtr = CMgaParser::ResolveClosure4Derivation;
		m_clVer = 4;
		m_clVerStr = "4";//end

		project_prefs = project_prefs_orig | MGAPREF_IGNORECONNCHECKS;
		COMTHROW( p->put_Preferences(project_prefs) );

		COMTHROW( resolver.CoCreateInstance(L"Mga.MgaResolver") );
		ASSERT( resolver != NULL );

		project = p;

		CopyTo(filename, xmlfile);

		try
		{
			XMLPlatformUtils::Initialize();

			SAXParser parser;
			parser.setDoValidation(true);
			parser.setDocumentHandler(this);
			parser.setErrorHandler(this);
			parser.setEntityResolver(this);

			elementfuncs = elementfuncs_scmga;

			// manual first pass

			pass_count = 1;

			ranges.clear();
			ranges.push_front();
			ranges.front().begin = 1;
			ranges.front().end = (counter_type)-1;
			ranges.front().previous.name = "start";
			ranges.front().previous.object = target;
			skip_element_level = 0;
			
			parser.parse(xmlfile.c_str());

			ASSERT( ranges.front().begin == 1 );
			ranges.pop_front();
			elements.clear();

			max_counter = counter;

			// the other passes

			parser.setDoValidation(false);

			while( !ranges.empty() && ranges.front().begin != (counter_type)-1 )
			{
				// FIXME: better algorithm for infinite loop
				if( ++pass_count >= 1 )//<!> reduced from 100 to 5
					HR_THROW(E_TOOMANYPASSES);

				parser.parse(xmlfile.c_str());
			}

			ASSERT( elements.empty() );
			ranges.clear();

			if( isNeedFor2ndStep())
			{
				CComBSTR bstr1("[Smartcopy Parser] Some objects were not found during parsing.");
				CComBSTR bstr2("[Smartcopy Parser] Invoking 2nd step...");
				if( m_GME) COMTHROW( m_GME->ConsoleMessage( bstr1, MSG_ERROR));
				if( m_GME) COMTHROW( m_GME->ConsoleMessage( bstr2, MSG_INFO));
				
				tryToFindMissedReferreds();
				tryToFindMissedSetMembers();
			}
		}
	    catch(const XMLException &e)
		{
			XmlStr desc(e.getMessage());

			ThrowXmlError(desc.c_str());
		}
		COMTHROW( project->put_Preferences(project_prefs_orig) );

		for(librecords::iterator i = libstodo.begin(); i != libstodo.end(); i++) { // copied from ParseProject in order to recognize libraries (zolmol)
			COMTHROW(i->f->put_LibraryName(CComBSTR(i->libname.c_str())));
			COMTHROW(i->f->put_Exempt(VARIANT_FALSE));
		}
		libstodo.clear();

		COMTHROW( project->CommitTransaction() );
		project = NULL;
		clear_GME( m_GME);

		CloseAll();
	}
	catch(hresult_exception &e)
	{
		CloseAll();

		ASSERT( FAILED(e.hr) );
		if( e.hr == E_XMLPARSER )
			SetErrorInfo(errorinfo);
		else
			SetErrorInfo2(e.hr);

		return e.hr;
	}
	return S_OK;
}

CMgaParser::elementfunc CMgaParser::elementfuncs_scmga[] = 
{
	//elementfunc("project", StartProject, EndNone),
	elementfunc("name", StartNone, EndName),
	elementfunc("comment", StartNone, EndComment),
	elementfunc("author", StartNone, EndAuthor),
	elementfunc("value", StartNone, EndValue),
	elementfunc("regnode", StartRegNode, EndNone),
	elementfunc("attribute", StartAttribute, EndNone),
	elementfunc("constraint", StartNone, EndConstraint),
	elementfunc("folder", StartFolder, EndNone),
	elementfunc("model", StartModel, EndNone),
	elementfunc("atom", StartAtom, EndNone),
	elementfunc("connection", StartSCConnection, EndSCConnection),	//
	elementfunc("connpoint", StartSCConnPoint, EndNone),	//
	elementfunc("reference", StartSCReference, EndSCReference),	//
	elementfunc("set", StartSCSet, EndSCSet),				//
	elementfunc("clipboard", StartBCClipboard, EndNone),	// common for closures
	elementfunc("", NULL, NULL)
};

void CMgaParser::ResolveClosure4Derivation(const attributes_type &attributes, deriv_type &deriv)
{
	const std::string *s = GetByNameX(attributes, "derivedfrom");
	if( s == NULL )
	{
		s = GetByNameX(attributes, "closurelibderivedfrom");
		if( s != NULL)
		{
			CComObjPtr<IMgaObject> obj;
			findObjOnAbsPath( project, *s, obj, "archetype");
			if( obj)
				COMTHROW( obj.QueryInterface( deriv.from));
			else if( m_GME)
			{
				CComBSTR bstr("[Smartcopy Parser] Archetype can not be found in library. Basetype lost. ");
				bstr.Append("Search path used: ");
				bstr.Append( makeViewable(*s).c_str());
				COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
			}

		}
		else //if( s == NULL)
		{
			s = GetByNameX( attributes, "closure2derivedfrom");
			if( s != NULL)
			{
				CComObjPtr<IMgaFolder> place_f;
				CComObjPtr<IMgaModel> place_m;
				//<!>
				// it searches for the objects place (one more time this search will occur later) 
				bool place_found = findPlaceForElem( attributes, deriv, place_f, place_m);
				
				std::string rel_path_changed = *s;
				// since the derfrom is calculated relatively to the fco, 
				// we have to modify the path so that it could be found
				// with a search starting from place_m _f
				if( rel_path_changed.substr(0, 4) == "/@..")
					rel_path_changed = rel_path_changed.substr(4);
				else if( rel_path_changed == "")
				{
					//<!>
					// the object has similar name to its archetype
					ASSERT(0);
					deriv.from.Release();
					return;
				}
				else // the derfrom object cannot be inside the derived object
				{
					ASSERT(0);
					deriv.from.Release();
					return;
				}

				CComObjPtr<IMgaObject> obj;
				if( place_f)
					findObjOnRelPath( CComObjPtr<IMgaObject>(place_f), rel_path_changed, obj, "archetype");
				else if(place_m)
					findObjOnRelPath( CComObjPtr<IMgaObject>(place_m), rel_path_changed, obj, "archetype");

				if( obj)
					COMTHROW( obj.QueryInterface( deriv.from));
				else if( m_GME) // error report
				{
					CComBSTR bstr("[Smartcopy Parser] Archetype can not be found. Basetype lost. ");
					bstr.Append("Search path used: ");
					bstr.Append( makeViewable(rel_path_changed).c_str());
					COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
				}
			}
			else
			{
				deriv.from.Release();
				return;
			}
		}


		if ( deriv.from == NULL) // do not throw exception in case of closurelibderivedfrom
		{
			deriv.from.Release();
			return;
		}
	}
	else
	{
		LookupByID(*s, deriv.from);

		if( deriv.from == NULL )
			throw pass_exception();
	}

	s = GetByNameX(attributes, "isinstance");
	deriv.isinstance = ( s != NULL && *s == "yes" ) ? VARIANT_TRUE : VARIANT_FALSE;

	s = GetByNameX(attributes, "isprimary");
	deriv.isprimary = ( s != NULL && *s == "no" ) ? false : true;
}

// ------- Element Handlers
void CMgaParser::StartSCConnection(const attributes_type &attributes)
{
	bool place_found = false;
	CComObjPtr<IMgaFolder> place_f;
	CComObjPtr<IMgaModel> place_m;

	CComObjPtr<IMgaFCO> conn;
	deriv_type deriv;
	//ResolveDerivation(attributes, deriv);
	(*this.*m_resolveDerFuncPtr)(attributes, deriv);

	if( !(place_found = findPlaceForElem( attributes, deriv, place_f, place_m)))
		throw pass_exception();//<!>

	if( place_f)
	{
		CComObjPtr<IMgaFolder> prev;
		prev = place_f;

		preparerelid(attributes);
		if( deriv.from != NULL )
		{
			COMTHROW( prev->DeriveRootObject(deriv.from, deriv.isinstance, PutOut(conn)) );
		}
		else
		{
			CComObjPtr<IMgaMetaFCO> meta;
			COMTHROW( resolver->get_KindByStr(prev, PutInBstrAttr(attributes, "kind"), 
				OBJTYPE_CONNECTION, PutOut(meta)) );
			ASSERT( meta != NULL );

			COMTHROW( prev->CreateRootObject(meta, PutOut(conn)) );
		}
		assignrelid(conn);
	}
	else
	{
		ASSERT( place_m);
		CComObjPtr<IMgaModel> prev;
		prev = place_m;

		CComObjPtr<IMgaMetaRole> role;
		COMTHROW( resolver->get_RoleByStr(prev, 
			PutInBstrAttr(attributes, "kind"), OBJTYPE_CONNECTION,
			PutInBstrAttr(attributes, "role"), NULL, PutOut(role)) );
		ASSERT( role != NULL );

		if( deriv.from != NULL )
		{
			CComObjPtr<IMgaConnection> derivedfrom;
			COMTHROW( deriv.from.QueryInterface(derivedfrom) );

			if( deriv.isprimary )
			{
				preparerelid(attributes);
				COMTHROW( prev->DeriveChildObject(derivedfrom, role, deriv.isinstance, PutOut(conn)) );
				assignrelid(conn);
			}
			else
			{
				COMTHROW( prev->get_ChildDerivedFrom(derivedfrom, PutOut(conn)) );
			}
		}
		else
		{
			preparerelid(attributes);
			COMTHROW( prev->CreateChildObject(role, PutOut(conn)) );
			assignrelid(conn);
		}
	}
	ASSERT( conn != NULL );
	GetCurrent().object = conn;

	if( GetByName(attributes, "isbound") == "yes" ) GetCurrent().exstrinfo = "skip";

	RegisterLookup(attributes, conn);
}

void CMgaParser::StartSCConnPoint(const attributes_type &attributes)
{
	ASSERT( GetPrevName() == "connection" );
	CComObjPtr<IMgaConnection> conn;
	GetPrevObj(conn);
	
	CComBstrObj ccname; // does not have "closurename" attribute, that is why we use the connection's name property
	COMTHROW( conn->get_Name( PutOut( ccname)));
	std::string viewable_name; 
	CopyTo( ccname, viewable_name);

	if( GetPrevious().exstrinfo == "skip" || GetByName(attributes, "isbound") == "yes" ) return;

	CComObjPtr<IMgaFCO> target;

	const std::string *closlibtg = GetByNameX(attributes, "closurelibtarget");
	if ( closlibtg != NULL)
	{
		CComObjPtr<IMgaObject> obj;
		findObjOnAbsPath( project, *closlibtg, obj, "connection target");
		if ( obj)
			COMTHROW( obj.QueryInterface( target));
		else // error report
		{
			CComBSTR bstr("[Smartcopy Parser] Connection \"");
			bstr.Append( viewable_name.c_str());
			bstr.Append("\": target not found in library. ");
			bstr.Append("Search path used: ");
			bstr.Append( makeViewable( *closlibtg).c_str());
			if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
		}
	}

	if( closlibtg == NULL || !target)
	{
		const std::string *clos2tg = GetByNameX(attributes, "closure2target");
		if( clos2tg != NULL)
		{
			CComObjPtr<IMgaObject> obj;
			findObjOnRelPath( CComObjPtr<IMgaObject>(conn), *clos2tg, obj, "connection target");
			if ( obj)
				COMTHROW( obj.QueryInterface( target));
			else // error report
			{
				CComBSTR bstr("[Smartcopy Parser] Connection \"");
				bstr.Append( viewable_name.c_str());
				bstr.Append("\": target not found. ");
				bstr.Append("Search path used: ");
				bstr.Append( makeViewable( *clos2tg).c_str());
				if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
			}
		}
	}

	// throw even if closurelibtarget present but not found the needed object
	if ( target == NULL)
		throw pass_exception();

	CComObjPtr<IMgaFCOs> coll;
	COMTHROW(coll.CoCreateInstance(L"Mga.MgaFCOs"));

	const std::string *s = GetByNameX(attributes, "closure2refs");
	if( s != NULL )
	{
		int pos = s->find_first_not_of(' ', 0);
		ASSERT( pos >= 0 );
		while( pos < (int) s->length() )//zolmol!
		{
			int pos2 = s->find_first_of(' ', pos);
			if( pos2 < 0 )
				pos2 = s->length();
			ASSERT( pos2 > pos );

			CComObjPtr<IMgaFCO> ref;
			//LookupByID(std::string(*s, pos, pos2-pos), ref);
			CComObjPtr<IMgaObject> obj;
			std::string ref_str(*s, pos, pos2-pos);
			findObjOnRelPath( CComObjPtr<IMgaObject>(conn), ref_str, obj, "reference chain");
			if ( obj)
				COMTHROW( obj.QueryInterface( ref));
			else // error report
			{
				CComBSTR bstr("[Smartcopy Parser] Reference not found involved in Connection \"");
				bstr.Append( viewable_name.c_str());
				bstr.Append("\". ");
				bstr.Append("Search path used: ");
				bstr.Append( makeViewable( ref_str).c_str());
				if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
			}

			if( ref == NULL )
				throw pass_exception();

			COMTHROW(coll->Append(ref));

			pos = pos2+1;
		}
	}

	CComObjPtr<IMgaConnPoint> connpoint;

	COMTHROW( conn->AddConnPoint(PutInBstr(GetByName(attributes, "role")), 0,	// FIXME: multiplicity
		target, coll, PutOut(connpoint)) );
}

void CMgaParser::StartSCReference(const attributes_type &attributes)
{
	bool place_found = false;
	CComObjPtr<IMgaFolder> place_f;
	CComObjPtr<IMgaModel> place_m;

	CComObjPtr<IMgaFCO> fco;

	deriv_type deriv;
	//ResolveDerivation(attributes, deriv);
	(*this.*m_resolveDerFuncPtr)(attributes, deriv);

	CComObjPtr<IMgaFCO> referred;

	const std::string *s;

	if( !(place_found = findPlaceForElem( attributes, deriv, place_f, place_m)))
		throw pass_exception();//<!>

	if( place_f)
	{
		CComObjPtr<IMgaFolder> prev;
		prev = place_f;

		preparerelid(attributes);
		if( deriv.from != NULL )
		{
			COMTHROW( prev->DeriveRootObject(deriv.from, deriv.isinstance, PutOut(fco)) );
		}
		else
		{
			CComObjPtr<IMgaMetaFCO> meta;
			COMTHROW( resolver->get_KindByStr(prev, PutInBstrAttr(attributes, "kind"), 
				OBJTYPE_REFERENCE, PutOut(meta)) );
			ASSERT( meta != NULL );

			COMTHROW( prev->CreateRootObject(meta, PutOut(fco)) );
		}
		assignrelid(fco);
	}
	else
	{
		ASSERT( place_m);
		CComObjPtr<IMgaModel> prev;
		prev = place_m;

		CComObjPtr<IMgaMetaRole> role;
		COMTHROW( resolver->get_RoleByStr(prev, 
			PutInBstrAttr(attributes, "kind"), OBJTYPE_REFERENCE,
			PutInBstrAttr(attributes, "role"), NULL, PutOut(role)) );
		ASSERT( role != NULL );

		if( deriv.from != NULL )
		{
			CComObjPtr<IMgaReference> derivedfrom;
			COMTHROW( deriv.from.QueryInterface(derivedfrom) );

			if( deriv.isprimary )
			{
				preparerelid(attributes);
				COMTHROW( prev->DeriveChildObject(derivedfrom, role, deriv.isinstance, PutOut(fco)) );
				assignrelid(fco);
			}
			else
			{
				COMTHROW( prev->get_ChildDerivedFrom(derivedfrom, PutOut(fco)) );
			}
		}
		else
		{
			preparerelid(attributes);
			COMTHROW( prev->CreateChildObject(role, PutOut(fco)) );
			assignrelid(fco);
		}
	}
	ASSERT( fco != NULL );
	//fco is the reference just created

	s = GetByNameX(attributes, "closurelibreferred");
	if( s != NULL)
	{
		CComObjPtr<IMgaObject> obj;
		findObjOnAbsPath( project, *s, obj, "referred object");

		if( obj)
			COMTHROW( obj.QueryInterface( referred));
		else // report error
		{
			CComBSTR bstr("[Smartcopy Parser] Reference \"");
			bstr.Append( makeNameViewable( GetByName( attributes, "closurename")).c_str());
			bstr.Append("\": target not found in library. ");
			bstr.Append("Search path used: ");
			bstr.Append( makeViewable( *s).c_str());
			if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
		}
	}
	
	if( s == NULL || !referred)
	{
		s = GetByNameX(attributes, "closure2referred");
		if( s != NULL && s->compare("") != 0)
		{
			CComObjPtr<IMgaObject> obj;
			findObjOnRelPath( CComObjPtr<IMgaObject>(fco), *s, obj, "referred object");

			if ( obj)
				COMTHROW( obj.QueryInterface( referred));
			else // report error
			{
				CComBSTR bstr("[Smartcopy Parser] Reference \"");
				bstr.Append( makeNameViewable( GetByName( attributes, "closurename")).c_str());
				bstr.Append("\": target not found. ");
				bstr.Append("Search path used: ");
				bstr.Append( makeViewable( *s).c_str());
				if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));

				// store data to process in 2nd step
				m_notFoundReferredObject[ fco ] = *s;
			}
		}
	}

	if( !(GetByName(attributes, "isbound") == "yes") && referred != NULL )
	{
		CComObjPtr<IMgaReference> ref;
		COMTHROW( fco.QueryInterface(ref) );

		COMTHROW( ref->put_Referred(referred) );
	}

	GetCurrent().object = fco;

	RegisterLookup(attributes, fco);
	pushInfo( place_f, place_m, fco, deriv.from);
}

void CMgaParser::StartSCSet(const attributes_type &attributes)
{
	bool place_found;
	CComObjPtr<IMgaFolder> place_f;
	CComObjPtr<IMgaModel> place_m;
	CComObjPtr<IMgaFCO> fco;

	deriv_type deriv;
	//ResolveDerivation(attributes, deriv);
	(*this.*m_resolveDerFuncPtr)(attributes, deriv);

	if( !( place_found = findPlaceForElem( attributes, deriv, place_f, place_m)))
		throw pass_exception();//<!>

	if( place_f)
	{
		CComObjPtr<IMgaFolder> prev;
		prev = place_f;

		preparerelid(attributes);
		if( deriv.from != NULL )
		{
			COMTHROW( prev->DeriveRootObject(deriv.from, deriv.isinstance, PutOut(fco)) );
		}
		else
		{
			CComObjPtr<IMgaMetaFCO> meta;
			COMTHROW( resolver->get_KindByStr(prev, PutInBstrAttr(attributes, "kind"), 
				OBJTYPE_SET, PutOut(meta)) );
			ASSERT( meta != NULL );

			COMTHROW( prev->CreateRootObject(meta, PutOut(fco)) );
		}
		assignrelid(fco);
	}
	else
	{
		ASSERT( place_m);
		CComObjPtr<IMgaModel> prev;
		prev = place_m;

		CComObjPtr<IMgaMetaRole> role;
		COMTHROW( resolver->get_RoleByStr(prev, 
			PutInBstrAttr(attributes, "kind"), OBJTYPE_SET,
			PutInBstrAttr(attributes, "role"), NULL, PutOut(role)) );
		ASSERT( role != NULL );

		if( deriv.from != NULL )
		{
			CComObjPtr<IMgaSet> derivedfrom;
			COMTHROW( deriv.from.QueryInterface(derivedfrom) );

			if( deriv.isprimary )
			{
				preparerelid(attributes);
				COMTHROW( prev->DeriveChildObject(derivedfrom, role, deriv.isinstance, PutOut(fco)) );
				assignrelid(fco);
			}
			else
			{
				preparerelid(attributes);
				COMTHROW( prev->get_ChildDerivedFrom(derivedfrom, PutOut(fco)) );
				assignrelid(fco);
			}
		}
		else
		{
			preparerelid(attributes);
			COMTHROW( prev->CreateChildObject(role, PutOut(fco)) );
			assignrelid(fco);
		}
	}
	ASSERT( fco != NULL );
	std::list< CComObjPtr<IMgaFCO> > members;//slist

	const std::string *s = GetByNameX(attributes, "closure2members");
	if( s != NULL)
	{
		int pos = s->find_first_not_of(' ', 0);
		ASSERT( pos >= 0 );
		while( pos < (int) s->length() )//zolmol!
		{
			int pos2 = s->find_first_of(' ', pos);
			if( pos2 < 0 )
				pos2 = s->length();
			ASSERT( pos2 > pos );

			std::string member_str(*s, pos, pos2-pos);
			CComObjPtr<IMgaObject> obj;
			findObjOnRelPath( CComObjPtr<IMgaObject>(fco), member_str, obj, "set member");

			if ( obj)
			{
				CComObjPtr<IMgaFCO> member;
				COMTHROW( obj.QueryInterface( member));
				members.push_front(member);
			}
			else // error report
			{
				CComBSTR bstr("[Smartcopy Parser] Set \"");
				bstr.Append( makeNameViewable( GetByName( attributes, "closurename")).c_str());
				bstr.Append("\": member not found. ");
				bstr.Append("Search path used: ");
				bstr.Append( makeViewable( member_str).c_str());
				if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));

				// insert data to process in 2nd step
				m_notFoundSetMembers[ fco ].push_back( member_str);
			}

			pos = pos2+1;
		}
	}

	CComObjPtr<IMgaSet> mgaset;
	COMTHROW( fco.QueryInterface(mgaset) );

	if( !(GetByName(attributes, "isbound") == "yes") ) {
		COMTHROW( mgaset->RemoveAll() ); //by ZolMol: if not bound then the members are different, remove the inherited members
		std::list< CComObjPtr<IMgaFCO> >::iterator i = members.begin();//slist
		while( i != members.end() )
		{
			COMTHROW( mgaset->AddMember(*i) );
			++i;
		}
	}
	GetCurrent().object = fco;

	RegisterLookup(attributes, fco);
	pushInfo( place_f, place_m, fco, deriv.from);
}



void CMgaParser::EndSCReference()
{
	popInfo();
}

void CMgaParser::EndSCSet()
{
	popInfo();
}

void CMgaParser::EndSCConnection()
{
	ASSERT( GetCurrent().object);
	if( !GetCurrent().object) return;

	CComObjPtr<IMgaSimpleConnection> cur_conn;

	COMTHROW( GetCurrent().object.QueryInterface( cur_conn));

	if( !cur_conn) return;

	CComObjPtr<IMgaModel> parent_m;
	CComObjPtr<IMgaFolder> parent_f;

	COMTHROW( cur_conn->get_ParentModel( PutOut( parent_m)));
	COMTHROW( cur_conn->get_ParentFolder( PutOut( parent_f)));

	CComObjPtr<IMgaObject> parent_obj;
	if( parent_f)
		parent_obj = parent_f;
	else if( parent_m)
		parent_obj = parent_m;

	if( !parent_obj) return;

	CComObjPtr<IMgaFCO> s,d;
	COMTHROW( cur_conn->get_Src( PutOut( s)));
	COMTHROW( cur_conn->get_Dst( PutOut( d)));

	// 3rd param: ignore its basetype, anyway it is a connection
	// 4th param: error msg to be displayed if (s && d) is false
	writeInfo( parent_obj, CComObjPtr<IMgaObject>( cur_conn), CComObjPtr<IMgaObject>(), s && d );
}

bool CMgaParser::isNeedFor2ndStep()
{
	return !m_notFoundReferredObject.empty() || !m_notFoundSetMembers.empty();
}

void CMgaParser::tryToFindMissedReferreds()
{
	std::map< CComObjPtr<IMgaFCO>, std::string >::iterator it = m_notFoundReferredObject.begin();
	std::map< CComObjPtr<IMgaFCO>, std::string >::iterator itend = m_notFoundReferredObject.end();
	for( ; it != itend; ++it)
	{
		if( !it->first) continue;
		bool error = false;

		CComObjPtr<IMgaObject> target;
		findObjOnRelPath( CComObjPtr<IMgaObject>( it->first), it->second, target, "referred object");
		if( target)
		{
			CComObjPtr<IMgaReference> ref;
			COMTHROW( it->first.QueryInterface( ref));
			CComObjPtr<IMgaFCO> fco_target;
			COMTHROW( target.QueryInterface( fco_target));

			if( ref && fco_target)
			{
				COMTHROW( ref->put_Referred( fco_target));

				CComBSTR id1, nm1, id2, nm2, bstr("[Smartcopy Parser] Reference <A HREF=\"mga:");
				COMTHROW( ref->get_ID( &id1));
				COMTHROW( ref->get_Name( &nm1));
				COMTHROW( fco_target->get_ID( &id2));
				COMTHROW( fco_target->get_Name( &nm2));

				bstr.AppendBSTR( id1);
				bstr.Append("\">");
				if( nm1.Length() != 0)
					bstr.AppendBSTR( nm1);
				else
					bstr.Append("emptyname");

				bstr.Append("</A> set to refer to <A HREF=\"mga:");
				bstr.AppendBSTR( id2);
				bstr.Append("\">");
				if( nm2.Length() != 0)
					bstr.AppendBSTR( nm2);
				else
					bstr.Append("emptyname");

				bstr.Append("</A> in 2nd step successfully.");
				if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_INFO));
			}
			else
				error = true;
		}
		else 
			error = true;

		if( error)
		{
			CComBSTR id1, nm1, bstr( "[Smartcopy Parser] Reference <A HREF=\"mga:");
			COMTHROW( it->first->get_ID( &id1));
			COMTHROW( it->first->get_Name( &nm1));
			bstr.AppendBSTR( id1);
			bstr.Append("\">");
			if( nm1.Length() != 0)
				bstr.AppendBSTR( nm1);
			else
				bstr.Append("emptyname");

			bstr.Append("</A>: target not found in 2nd step. ");
			bstr.Append("Search path used: ");
			bstr.Append( makeViewable( it->second).c_str());

			if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
		}
	}
}

void CMgaParser::tryToFindMissedSetMembers()
{
	std::map< CComObjPtr<IMgaFCO>, std::vector< std::string >, CompareCComObj >::iterator it = m_notFoundSetMembers.begin();
	std::map< CComObjPtr<IMgaFCO>, std::vector< std::string >, CompareCComObj >::iterator itend = m_notFoundSetMembers.end();
	for( ; it != itend; ++it)
	{
		if( !it->first || it->second.empty()) continue;

		CComObjPtr<IMgaSet> set;
		COMTHROW( it->first.QueryInterface( set));

		if( !set) continue;

		std::vector< std::string >::iterator member_it = it->second.begin();
		std::vector< std::string >::iterator member_it_end = it->second.end();
		for( ; member_it != member_it_end; ++member_it)
		{
			if( member_it->empty()) continue;

			bool error = false;
			CComObjPtr<IMgaObject> member;
			findObjOnRelPath( CComObjPtr<IMgaObject>( it->first), *member_it, member, "set member");
			if( member)
			{
				CComObjPtr<IMgaFCO> fco_member;
				COMTHROW( member.QueryInterface( fco_member));

				if( fco_member)
				{
					COMTHROW( set->AddMember( fco_member));

					CComBSTR id1, nm1, id2, nm2, bstr("[Smartcopy Parser] Member <A HREF=\"mga:");
					COMTHROW( set->get_ID( &id1));
					COMTHROW( set->get_Name( &nm1));
					COMTHROW( fco_member->get_ID( &id2));
					COMTHROW( fco_member->get_Name( &nm2));

					bstr.AppendBSTR( id2);
					bstr.Append("\">");
					if( nm2.Length() != 0)
						bstr.AppendBSTR( nm2);
					else
						bstr.Append("emptyname");

					bstr.Append("</A> added to set <A HREF=\"mga:");
					bstr.AppendBSTR( id1);
					bstr.Append("\">");
					if( nm1.Length() != 0)
						bstr.AppendBSTR( nm1);
					else
						bstr.Append("emptyname");

					bstr.Append("</A> in 2nd step successfully.");
					if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_INFO));
				}
				else
					error = true;
			}
			else
				error = true;

			if( error)
			{
				CComBSTR id1, nm1, bstr( "[Smartcopy Parser] Set <A HREF=\"mga:");
				COMTHROW( set->get_ID( &id1));
				COMTHROW( set->get_Name( &nm1));
				bstr.AppendBSTR( id1);
				bstr.Append("\">");
				if( nm1.Length() != 0)
					bstr.AppendBSTR( nm1);
				else
					bstr.Append("emptyname");

				bstr.Append("</A>: member not found in 2nd step. ");
				bstr.Append("Search path used: ");
				bstr.Append( makeViewable( *member_it).c_str());

				if( m_GME) COMTHROW( m_GME->ConsoleMessage(bstr, MSG_ERROR));
			}
		}
	}
}