#include "stdafx.h"
#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include "CoreXmlFile.h"
#include "CommonCollection.h"
#include "VSSLoginDlg.h"
#include "FilesInUseDlg.h"
#include "FilesInUseDetailsDlg.h"

using namespace XERCES_CPP_NAMESPACE;
using std::string;

void encodeXMLString( string& in, string& out )
{
    int size = in.length();
    out = "";

    for( int i=0; i<size; ++i )
    {
        switch( in[i] )
        {
            case '<' : out += "&lt;"; break;
            case '>' : out += "&gt;"; break;
            case '&' : out += "&amp;"; break;
            case '\'': out += "&apos;"; break;
            case '\"': out += "&quot;"; break;
            default  : out += in[i];
        }
    }
}


void bin2string( const unsigned char * bytes, int len, std::string& str )
{
    char hex[3];
    str.clear();
    for( int i=0; i<len; ++i )
    {
        sprintf( hex, "%02x", bytes[i] );
        str += hex;
    }
}

unsigned char hexCharToInt( char ch )
{
    if( ch>='0' && ch<='9' )
        return ch - '0';

    if( ch == 'a' || ch == 'A' )
        return 10;
    if( ch == 'b' || ch == 'B' )
        return 11;
    if( ch == 'c' || ch == 'C' )
        return 12;
    if( ch == 'd' || ch == 'D' )
        return 13;
    if( ch == 'e' || ch == 'E' )
        return 14;
    if( ch == 'f' || ch == 'F' )
        return 15;
    ASSERT(false);
    return 0;
}

void string2bin( const char * str, unsigned char * bytes )
{
    int l = strlen(str) / 2;
    for( int i=0; i<l; ++i )
    {        
        unsigned char c1 = hexCharToInt( str[i*2] );
        unsigned char c2 = hexCharToInt( str[i*2 + 1] );
        //unsigned char c;
        //sscanf( str + i*2, "%02x", &c );      // do not work in release veriosn...
        bytes[i] = 16 * c1 + c2;
    }
}

void guid2str( GUID guid, std::string& str )
{
    if( guid == GUID_NULL )
        str = "null";
    else
        bin2string( (unsigned char*)&guid, sizeof(GUID), str );
}

GUID str2guid( const char * str )
{    
    int len = strlen(str);
    if( len == 0 || strcmp( str, "null" ) == 0 )
        return GUID_NULL;
    else
    {
        GUID guid;
        string2bin( str, (unsigned char*)&guid );
        return guid;
    }
}

void replaceSpaceWithUnderscore( std::string& str )
{
    for( unsigned int i=0; i < str.size(); ++i )
        if( str[i] == ' ' )
            str[i] = '_';
}

void replaceUnderscoreWithSpace( char * str )
{
    if( str == NULL )
        return;

    int n = strlen(str);
    for( int i=0; i<n; ++i )
        if( str[i] == '_' )
            str[i] = ' ';
}


////////////////////////////////////////////////////////////////////////////////
// XmlAttrBase class
////////////////////////////////////////////////////////////////////////////////

XmlAttrBase * XmlAttrBase::create(valtype_type valtype)
{
    ASSERT( valtype != VALTYPE_NONE );

	XmlAttrBase * xmlattr = NULL;

	switch(valtype)
	{
    case VALTYPE_LONG:
		xmlattr = new XmlAttrLong();
		break;

	case VALTYPE_STRING:
		xmlattr = new XmlAttrString();
		break;

	case VALTYPE_BINARY:
		xmlattr = new XmlAttrBinary();
		break;

	case VALTYPE_LOCK:
		xmlattr = new XmlAttrLock();
		break;

	case VALTYPE_POINTER:
		xmlattr = new XmlAttrPointer();
		break;

	case VALTYPE_COLLECTION:
		xmlattr = new XmlAttrCollection();
		break;

	case VALTYPE_REAL:
		xmlattr = new XmlAttrReal();
		break;

	default:
		HR_THROW(E_METAPROJECT);
	}

	if( xmlattr == NULL )
		HR_THROW(E_OUTOFMEMORY);

	return xmlattr;
}

XmlAttrBase::XmlAttrBase()
{ 
}
    
XmlAttrBase::~XmlAttrBase() 
{ 
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrLong class
////////////////////////////////////////////////////////////////////////////////

XmlAttrLong::XmlAttrLong()
{
    m_value = 0;
}

valtype_type XmlAttrLong::getType() const
{
    return VALTYPE_LONG;
}

void XmlAttrLong::fromVariant(VARIANT p)
{
    CopyTo(p, m_value);
}

void XmlAttrLong::toVariant(VARIANT *p) const
{
    CopyTo(m_value, p);
}

void XmlAttrLong::fromString(const char * str)
{
    if( str == NULL || strlen(str)==0 )
        m_value = 0;
    else
        m_value = atoi( str );
}

void XmlAttrLong::toString(std::string& str) const
{
    static char buf[100];
    sprintf( buf, "%d", m_value );
    str = buf;
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrReal class
////////////////////////////////////////////////////////////////////////////////

XmlAttrReal::XmlAttrReal()
{
    m_value = 0;
}

valtype_type XmlAttrReal::getType() const
{
    return VALTYPE_REAL;
}

void XmlAttrReal::fromVariant(VARIANT p)
{
    CopyTo(p, m_value);
}

void XmlAttrReal::toVariant(VARIANT *p) const
{
    CopyTo(m_value, p);
}

void XmlAttrReal::fromString(const char * str)
{
    if( str == NULL || strlen(str)==0 )
        m_value = 0;
    else
        m_value = atof( str );
}

void XmlAttrReal::toString(std::string& str) const
{
    static char buf[100];
    sprintf( buf, "%.6f", m_value );
    str = buf;
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrString class
////////////////////////////////////////////////////////////////////////////////

valtype_type XmlAttrString::getType() const
{
    return VALTYPE_STRING;
}

void XmlAttrString::fromVariant(VARIANT p)
{
    CopyTo(p, m_value);
}

void XmlAttrString::toVariant(VARIANT *p) const
{
    CopyTo(m_value, p);
}

void XmlAttrString::fromString(const char * str)
{
    if( str == NULL )
        m_value = "";
    else
        m_value = str;
}

void XmlAttrString::toString(std::string& str) const
{
    str = m_value;
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrBinary class
////////////////////////////////////////////////////////////////////////////////

valtype_type XmlAttrBinary::getType() const
{
    return VALTYPE_BINARY;
}

void XmlAttrBinary::fromVariant(VARIANT p)
{
    CopyTo(p, m_value);
}

void XmlAttrBinary::toVariant(VARIANT *p) const
{
    CopyTo(m_value, p);
}

void XmlAttrBinary::fromString(const char * str)
{   
    if( str == NULL || strlen(str) == 0 )
        m_value.clear();
    else
    {
        // TODO: optimize this code
        int             len  = strlen(str)/2;
        unsigned char * buff = new unsigned char[len];
        string2bin( str, buff );
        m_value.resize( len );
        for( int i=0; i<len; ++i )
            m_value [i] = buff[i];
        delete [] buff;
    }
}

void XmlAttrBinary::toString(std::string& str) const
{
    bin2string( m_value.begin(), m_value.size(), str );
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrLock class
////////////////////////////////////////////////////////////////////////////////

XmlAttrLock::XmlAttrLock()
{
    m_value = LOCKING_NONE;
}

valtype_type XmlAttrLock::getType() const
{
    return VALTYPE_LOCK;
}

void XmlAttrLock::fromVariant(VARIANT p)
{
    CopyTo(p, m_value);
}

void XmlAttrLock::toVariant(VARIANT *p) const
{
    CopyTo(m_value, p);
}

void XmlAttrLock::fromString(const char * str)
{   
    if( str == NULL || strlen(str) == 0 )
        m_value = LOCKING_NONE;
    else
        m_value = atoi( str );
}

void XmlAttrLock::toString(std::string& str) const
{
    static char buf[100];
    sprintf( buf, "%d", m_value );
    str = buf;
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrPointer class
////////////////////////////////////////////////////////////////////////////////

XmlAttrPointer::XmlAttrPointer()
{
    m_parent = NULL;
}

valtype_type XmlAttrPointer::getType() const
{
    return VALTYPE_POINTER;
}

////////////////////////////////////////////////////////////////////////////////
// XmlAttrCollection class
////////////////////////////////////////////////////////////////////////////////

valtype_type XmlAttrCollection::getType() const
{
    return VALTYPE_COLLECTION;
}

////////////////////////////////////////////////////////////////////////////////
// XmlObject class
////////////////////////////////////////////////////////////////////////////////

XmlObject::XmlObject(ICoreMetaObject *metaobject, bool createAllAttributes )
{
    m_deleted         = false;
    m_modified        = false;
    m_loaded          = createAllAttributes;
    
    COMTHROW( metaobject->get_MetaID( &m_metaid ) );
    if( m_loaded )
        createAttributes(metaobject,ATTR_ALL);
    else
        createAttributes(metaobject,ATTR_PRIMARY);
    CoCreateGuid(&m_guid);

    m_sourceControlInfoRegNode = false;
}

XmlObject::~XmlObject()
{
    for( AttribMapIter it = m_attributes.begin(); it != m_attributes.end(); ++it )
        delete it->second;
}

bool XmlObject::isContainer()
{
    return( m_metaid==METAID_ROOT || m_metaid==DTID_MODEL || m_metaid==DTID_FOLDER );
}


void XmlObject::createAttributes(ICoreMetaObject *metaobject, int attrSet )
{
    ASSERT( metaobject != NULL );
    ASSERT( attrSet>=ATTR_PRIMARY && attrSet<=ATTR_ALL );

	CComObjPtr<ICoreMetaAttributes> metaattributes;
	COMTHROW( metaobject->get_Attributes(PutOut(metaattributes)) );
	ASSERT( metaattributes != NULL );

	typedef std::vector< CComObjPtr<ICoreMetaAttribute> > metaattributelist_type;
	metaattributelist_type metaattributelist;
	GetAll<ICoreMetaAttributes, ICoreMetaAttribute>(metaattributes, metaattributelist);

    for( metaattributelist_type::iterator i=metaattributelist.begin(); i!=metaattributelist.end(); ++i )
    {
		valtype_type valtype;
		COMTHROW( (*i)->get_ValueType(&valtype) );
        if( attrSet==ATTR_ALL ||
           (attrSet==ATTR_PRIMARY && (valtype==VALTYPE_POINTER || valtype==VALTYPE_COLLECTION || valtype==VALTYPE_LOCK)) ||
           (attrSet==ATTR_SECONDARY && (valtype!=VALTYPE_LOCK && valtype!=VALTYPE_POINTER && valtype!=VALTYPE_COLLECTION)))
        {
		    attrid_type attrId = ATTRID_NONE;
		    COMTHROW( (*i)->get_AttrID(&attrId) );
            ASSERT( attrId != ATTRID_NONE );
            m_attributes.insert( AttribMap::value_type(attrId,XmlAttrBase::create(valtype)) );
        }
	}
}

void XmlObject::deleteSecondaryAttribs()
{
    for( AttribMapIter it = m_attributes.begin(); it != m_attributes.end(); ++it )
    {
        valtype_type type =it->second->getType();
        if( type!=VALTYPE_LOCK && type!=VALTYPE_POINTER && type!=VALTYPE_COLLECTION )
        {
            delete it->second;
            m_attributes.erase(it);
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
// CCoreXmlFile class
////////////////////////////////////////////////////////////////////////////////

CCoreXmlFile::CCoreXmlFile()
{
    m_opened                = false;
    m_inTransaction         = false;
    m_modified              = false;
    m_metaAttributeId       = ATTRID_NONE;
    m_metaAttributeValType  = VALTYPE_NONE;
    m_openedObject          = NULL;
    m_sourceControl         = SC_NONE;
    m_savedOnce             = false;

    fillParentMap();

    XMLPlatformUtils::Initialize();
}

CCoreXmlFile::~CCoreXmlFile()
{
    clearAll();
}

STDMETHODIMP CCoreXmlFile::get_MetaProject(ICoreMetaProject **p)
{
    CHECK_OUT(p);
	CopyTo(m_metaProject, p);
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::put_MetaProject(ICoreMetaProject *p)
{
    COMTRY
	{
		closeMetaProject();
		m_metaProject = p;
	}
	COMCATCH( closeMetaProject() )
}

STDMETHODIMP CCoreXmlFile::get_MetaObject(ICoreMetaObject **p)
{
    CHECK_OUT(p);
	CopyTo(m_metaObject, p);
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::put_MetaObject(ICoreMetaObject *p)
{
    if( m_metaProject == NULL )
		COMRETURN(E_INVALID_USAGE);

	if( m_metaObject == p )
		return S_OK;

	COMTRY
	{
		if( p != NULL )
		{
            // check the metaproject of the object, it must be the same as the metaproject
            // of the storage
			CComObjPtr<ICoreMetaProject> t;
			COMTHROW( p->get_Project(PutOut(t)) );
			if( !IsEqualObject(m_metaProject, t) )
				HR_THROW(E_SAMEPROJECT);
		}

		closeMetaObject();
		m_metaObject = p;
		if( m_metaObject != NULL )
			openMetaObject();
	}
	COMCATCH( closeMetaObject() )
}

STDMETHODIMP CCoreXmlFile::get_MetaID(metaid_type *p)
{
    CHECK_OUT(p);
	*p = m_metaObjectId;
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::put_MetaID(metaid_type metaid)
{
    if( m_metaProject == NULL )
		COMRETURN(E_INVALID_USAGE);

	COMTRY
	{
		if( metaid != METAID_NONE )
		{
			CComObjPtr<ICoreMetaObject> p;
			COMTHROW( m_metaProject->get_Object(metaid, PutOut(p)) );
			ASSERT( p != NULL );

			if( m_metaObject != p )
			{
				closeMetaObject();
				MoveTo(p, m_metaObject);
				openMetaObject();
			}
		}
		else
			closeMetaObject();
	}
	COMCATCH( closeMetaObject() )
}

STDMETHODIMP CCoreXmlFile::get_MetaAttribute(ICoreMetaAttribute **p)
{
	CHECK_OUT(p);
	CopyTo(m_metaAttribute, p);
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::put_MetaAttribute(ICoreMetaAttribute *p)
{
    if( m_metaObject == NULL )
		COMRETURN(E_INVALID_USAGE);
	ASSERT( m_metaProject != NULL );

	if( m_metaAttribute == p )
		return S_OK;

	COMTRY
	{
		if( m_metaAttribute != NULL )
		{
            // the metaobject of the given metaattribute must be the same as the metaobject of 
            // the storage
			CComObjPtr<ICoreMetaObject> t;
			COMTHROW( m_metaAttribute->get_Object(PutOut(t)) );
			if( !IsEqualObject(m_metaObject, t) )
			{
				m_metaAttribute = NULL;
				return E_INVALIDARG;
			}
		}

		closeMetaAttribute();
		m_metaAttribute = p;
		if( m_metaAttribute != NULL )
			openMetaAttribute();
	}
	COMCATCH( closeMetaAttribute() )
}

STDMETHODIMP CCoreXmlFile::get_AttrID(attrid_type *p)
{
	CHECK_OUT(p);
	if( m_metaAttribute )
		return m_metaAttribute->get_AttrID(p);
	*p = 0;
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::put_AttrID(attrid_type attrid)
{
    if( m_metaObject == NULL )
		COMRETURN(E_INVALID_USAGE);
	ASSERT( m_metaProject != NULL );

	COMTRY
	{
		if( attrid != ATTRID_NONE )
		{
			CComObjPtr<ICoreMetaAttribute> p;
			COMTHROW( m_metaObject->get_Attribute(attrid, PutOut(p)) );
			ASSERT( p != NULL );

			if( m_metaAttribute != p )
			{
				closeMetaAttribute();
				m_metaAttribute = p;
				openMetaAttribute();
			}
		}
		else
			closeMetaAttribute();
	}
	COMCATCH( closeMetaAttribute() )
}

STDMETHODIMP CCoreXmlFile::get_AttributeValue(VARIANT *p)
{
    CHECK_OUT(p);

	if( m_openedObject == NULL || !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);
	
	COMTRY
	{
        AttribMapIter it = m_openedObject->m_attributes.find( m_metaAttributeId );
        if( m_metaAttributeValType == VALTYPE_POINTER )
        {
            getPointer( (XmlAttrPointer*)it->second, p );
        }
        else if( m_metaAttributeValType == VALTYPE_COLLECTION )
        {
            getCollection( (XmlAttrCollection*)it->second, p );
        }
        else if( m_metaAttributeValType == VALTYPE_LOCK )
        {
            it->second->toVariant(p);
        }
        else
        {
            if( !m_openedObject->m_loaded )
            {
                fullReadContainer(getContainer(m_openedObject));
                it = m_openedObject->m_attributes.find( m_metaAttributeId );
            }

            it->second->toVariant(p);
        }
	}
	COMCATCH(;)
}

STDMETHODIMP CCoreXmlFile::put_AttributeValue(VARIANT p)
{
    if( m_openedObject == NULL || !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);

    // root's guid changes evry time the project is modified we should igonre it
    // it casues some meta interpreter problems!
    if( m_metaAttributeId == ATTRID_GUID )
        return S_OK;

    // setting root's ATTRID_MDATE to "updatesourcecontrolinfo" means we have to update sourcecontrolinfo
    // and ignore attribute setting
    if( m_metaAttributeId == ATTRID_MDATE )
    {
        string str;
        CopyTo(p, str);
        if( strcmp( str.c_str(), "UpdateSourceControlInfo" ) == 0 )
        {
            updateSourceControlInfo();
            return S_OK;
        }
    }

    /*if( m_metaAttributeId == ATTRID_MGAVERSION )
    {
        long gmepointer;
        CopyTo(p, gmepointer);
        m_gme = (IGMEOLEApp*)gmepointer;        
        return S_OK;
    }*/

    // ignore UpdateSourceControlInfo regnode
    if( m_openedObject->m_sourceControlInfoRegNode && m_metaAttributeId == ATTRID_REGNODEVALUE)
        return S_OK;

    
    // TODO: return with specific error code
    //if( m_openedObject->m_readOnly )
      //  COMRETURN(E_INVALID_USAGE);
	
	COMTRY
	{
        AttribMapIter it = m_openedObject->m_attributes.find( m_metaAttributeId );
        if( m_metaAttributeValType == VALTYPE_POINTER )
        {   
            m_modifiedObjects.insert( m_openedObject );

            // handle special case: if this is the containment parent and this object is an inner
            // we have to check out the parent too
            ParentMap::iterator it2 = m_parentMap.find( m_openedObject->m_metaid );
            ASSERT( it2 != m_parentMap.end() );
            if( it2->second == m_metaAttributeId && !m_openedObject->isContainer() )
            {
                XmlObject * cont = getContainer(m_openedObject);
                m_modifiedObjects.insert( cont );
            }
            
            XmlAttrPointer * pointerAttr = (XmlAttrPointer*)it->second;
            if( m_undoMap.find(it->second) == m_undoMap.end() )
            {
                std::pair<UndoMapIter, bool> t = m_undoMap.insert( UndoMap::value_type(it->second,UndoItem(m_openedObject, m_metaAttributeId, CComVariant()) ));
                getPointer( pointerAttr, PutOut(t.first->second.m_value) );
            }
            setPointer( p );
        }
        else if( m_metaAttributeValType == VALTYPE_COLLECTION )
        {
            ASSERT( false );
        }
        else if( m_metaAttributeValType == VALTYPE_LOCK )
        {
            if( m_undoMap.find(it->second) == m_undoMap.end() )
            {
                std::pair<UndoMapIter, bool> t = m_undoMap.insert( UndoMap::value_type(it->second,UndoItem(m_openedObject, m_metaAttributeId, CComVariant()) ));
                it->second->toVariant( PutOut(t.first->second.m_value) );
            }

            it->second->fromVariant(p);

            //TODO: ha a lock 0-ra megy akkor ki lehet irni file-ba
        }
        else
        {
            m_modifiedObjects.insert( m_openedObject );

            if( !m_openedObject->m_loaded )
            {
                fullReadContainer(getContainer(m_openedObject));
                it = m_openedObject->m_attributes.find( m_metaAttributeId );
            }

            // save previous value to m_undoMap (if this is the first modification)
            if( m_undoMap.find(it->second) == m_undoMap.end() )
            {
                std::pair<UndoMapIter, bool> t = m_undoMap.insert( UndoMap::value_type(it->second,UndoItem(m_openedObject, m_metaAttributeId, CComVariant()) ));
                it->second->toVariant( PutOut(t.first->second.m_value) );
            }
            it->second->fromVariant(p);
        }

        m_openedObject->m_modified = true;
        m_modified = true;
	}
	COMCATCH(;)
}

STDMETHODIMP CCoreXmlFile::OpenObject(objid_type objid)
{
    if( m_metaObject == NULL || !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);
	   
    metaobjidpair_type idpair;
	idpair.metaid = m_metaObjectId;
	idpair.objid  = objid;

	COMTRY
	{
        m_openedObject = objectFromObjId(idpair);

        if( m_openedObject->m_deleted )
            m_openedObject = NULL;
	}
	COMCATCH(;)
}

STDMETHODIMP CCoreXmlFile::CreateObject(objid_type *objid)
{
    CHECK_OUT(objid);

	if( m_metaObject == NULL || !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);

    ASSERT( m_metaObjectId != METAID_ROOT );

	COMTRY
	{				
        // create and add new object
        XmlObject * obj = new XmlObject(m_metaObject,true);
        addObject( obj );

        m_createdObjects.push_back(obj);

        m_modified = true;
        m_openedObject = obj;
		*objid = (long)obj;

        if( obj->m_metaid==DTID_MODEL || obj->m_metaid==DTID_FOLDER )
            createSourceControlInfoRegNode( obj );
	}
	COMCATCH(;)
}

STDMETHODIMP CCoreXmlFile::CloseObject()
{
    m_openedObject = NULL;
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::LockObject()
{
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::DeleteObject()
{
    if( m_openedObject == NULL || !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);

    m_openedObject->m_deleted = true;
    m_deletedObjects.insert( m_openedObject );

    m_modified = true;

    //m_modifiedObjects.insert( m_openedObject );
    // TODO: add container 


    CloseObject();
    
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::OpenProject(BSTR connection, VARIANT_BOOL *ro_mode)
{
    if( m_opened || m_metaProject == NULL )
        COMRETURN(E_INVALID_USAGE);

    AFX_MANAGE_STATE(AfxGetStaticModuleState( ));

    parseConnectionString( connection );
    setFileNames();
    readProjectFile();

    //if( !readBinaryCache() )
    if( true )
    {                
        // binary cache is not found, get latest and read all
        if( m_sourceControl != SC_NONE )
            getLatestVersion();        

        readAll( true );        
        //writeBinaryCache();
    }
    else
    {
        // graph has been read  successfully, get latest and update graph
        if( m_sourceControl != SC_NONE )
            getLatestAndLoad();
    }

    m_opened    = true;
    m_modified  = false;
    m_savedOnce = true;
    
    if(ro_mode!=NULL) 
        *ro_mode = VARIANT_FALSE;

    return S_OK;
}

STDMETHODIMP CCoreXmlFile::CreateProject(BSTR connection)
{
    if( m_opened || m_metaProject == NULL )
		COMRETURN(E_INVALID_USAGE);

    parseConnectionString( connection );
    setFileNames();
            
    // clear data structures
    clearAll();

    // query the metaobject for the root
    CComObjPtr<ICoreMetaObject> mo;
	COMTHROW( m_metaProject->get_Object(METAID_ROOT, PutOut(mo)) );
	ASSERT( mo != NULL );

    // create the root
    m_root = new XmlObject(mo,true);
    addObject( m_root );

    if( m_clearCaseString.size() > 0 )
    {
        // check if the folder under ClearCase source control
        if( createClearCaseApp() )
        {
            if( !isPathUnderClearCase( m_parentFolderPath.c_str() ))
            {
                AfxMessageBox( "This given project folder is not under ClearCase source control!" );
            }
            else
            {
                m_sourceControl = SC_CLEARCASE;
            }
        }
    }
    else
    {        
        if( m_vssDatabaseStr.size() > 0 )
            createSourceSafeDatabase();
    }
    
    createProjectFile();
    
    m_opened   = true;
    m_modified = false;
   
    return S_OK;
}

STDMETHODIMP CCoreXmlFile::SaveProject(BSTR connection, VARIANT_BOOL keepoldname = VARIANT_TRUE) 
{
    //writeBinaryCache();
    writeAll();

    if( m_sourceControl != SC_NONE )
    {
        if( !m_savedOnce )
            checkInAll(true);
        else
            checkInAll();

    }

    m_modified = false;
    m_savedOnce = true;

    return S_OK;
}

STDMETHODIMP CCoreXmlFile::CloseProject( VARIANT_BOOL abort)
{
    if( !m_opened || m_metaProject == NULL )
		COMRETURN(E_INVALID_USAGE);

	COMTRY
	{
		/*if( abort == VARIANT_FALSE && m_modified ) 
			SaveProject(NULL);

        if( m_sourceControl != SC_NONE )
            checkInAll();*/

        clearAll();
	}
	COMCATCH(;)

	return S_OK;
}

STDMETHODIMP CCoreXmlFile::DeleteProject()
{ 
    return E_NOTIMPL;
}

STDMETHODIMP CCoreXmlFile::BeginTransaction()
{	
    if( !m_opened || m_inTransaction )
		COMRETURN(E_INVALID_USAGE);
	m_inTransaction = true;
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::CommitTransaction()
{    
    if( !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);

	ASSERT( m_opened );

    bool failed = false;
    
    XmlObjSet checkOutContainers;
    getCheckOutContainers(m_modifiedObjects, checkOutContainers);

    CloseObject();

    //m_needClose = false;

    if( !checkOutFiles(checkOutContainers) )
    {
        return E_FAIL;
    }
    else
    {
        m_createdObjects.clear();
        m_deletedObjects.clear();
        m_modifiedObjects.clear();
        m_undoMap.clear();
        m_inTransaction = false;


        //if( m_needClose )
          //  m_gme->CloseProject( TRUE );

        return S_OK;
    }
}

STDMETHODIMP CCoreXmlFile::AbortTransaction()
{
	if( !m_inTransaction )
		COMRETURN(E_INVALID_USAGE);

	ASSERT( m_opened );
    
    // undelete deleted objects
    for( XmlObjSetIter it=m_deletedObjects.begin(); it!=m_deletedObjects.end(); ++it )
        (*it)->m_deleted = false;

    // delete created objects
    int minIndex = m_objects.size();
    for( XmlObjVecIter it2=m_createdObjects.begin(); it2!=m_createdObjects.end(); ++it2 )
    {
        XmlObject* obj = *it2;
        m_objectsByGUID.erase(obj->m_guid);
        if( obj->m_index < minIndex )
            minIndex = obj->m_index;
    }
    m_objects.resize( minIndex );

    // rollback attributum changes
    for( UndoMapIter it3=m_undoMap.begin(); it3!=m_undoMap.end(); ++it3 )
    {
        XmlObject * obj = it3->second.m_object;
        if( it3->first->getType() == VALTYPE_POINTER )
        {
            XmlAttrPointer * pointerAttr = (XmlAttrPointer*)it3->first;
            metaobjidpair_type idpair;
            CopyTo(it3->second.m_value, idpair);
            XmlObject * parent = objectFromObjId(idpair);
            setPointer( obj, it3->second.m_attrId, parent );
        }
        else
        {
            it3->first->fromVariant(it3->second.m_value);
        }
    }
    m_undoMap.clear();

    m_deletedObjects.clear();
    m_createdObjects.clear();
    m_modifiedObjects.clear();
	CloseObject();	
	m_inTransaction = false;
	return S_OK;
}

STDMETHODIMP CCoreXmlFile::get_StorageType(long *p)
{
    CHECK_OUT(p);
	*p = 0;
	return S_OK;
}

void CCoreXmlFile::fillParentMap()
{
    m_parentMap.clear();

    m_parentMap.insert( ParentMap::value_type( DTID_CONSTRAINT , ATTRID_CONSTROWNER ) );
    m_parentMap.insert( ParentMap::value_type( DTID_REGNODE    , ATTRID_REGNOWNER   ) );
    m_parentMap.insert( ParentMap::value_type( DTID_FOLDER     , ATTRID_PARENT      ) );
    m_parentMap.insert( ParentMap::value_type( DTID_MODEL      , ATTRID_PARENT      ) );
    m_parentMap.insert( ParentMap::value_type( DTID_ATOM       , ATTRID_PARENT      ) );
    m_parentMap.insert( ParentMap::value_type( DTID_REFERENCE  , ATTRID_PARENT      ) );
    m_parentMap.insert( ParentMap::value_type( DTID_SET        , ATTRID_PARENT      ) );
    m_parentMap.insert( ParentMap::value_type( DTID_SETNODE    , ATTRID_SETMEMBER   ) );
    m_parentMap.insert( ParentMap::value_type( DTID_CONNECTION , ATTRID_PARENT      ) );
    m_parentMap.insert( ParentMap::value_type( DTID_CONNROLE   , ATTRID_CONNROLE    ) );
    m_parentMap.insert( ParentMap::value_type( DTID_CONNROLESEG, ATTRID_CONNSEG     ) );
    m_parentMap.insert( ParentMap::value_type( DTID_STRATTR    , ATTRID_ATTRPARENT  ) );
    m_parentMap.insert( ParentMap::value_type( DTID_INTATTR    , ATTRID_ATTRPARENT  ) );
    m_parentMap.insert( ParentMap::value_type( DTID_FLOATATTR  , ATTRID_ATTRPARENT  ) );
    m_parentMap.insert( ParentMap::value_type( DTID_BOOLATTR   , ATTRID_ATTRPARENT  ) );
    m_parentMap.insert( ParentMap::value_type( DTID_REFATTR    , ATTRID_ATTRPARENT  ) );
}

void CCoreXmlFile::closeMetaProject()
{
    closeMetaObject();
    CloseProject();
	m_metaProject = NULL;
}

void CCoreXmlFile::openMetaObject()
{
	ASSERT( m_metaObject != NULL );
	COMTHROW( m_metaObject->get_MetaID(&m_metaObjectId) );
}

void CCoreXmlFile::closeMetaObject()
{
	CloseObject();
	closeMetaAttribute();
	m_metaObject   = NULL;
	m_metaObjectId = METAID_NONE;
}

void CCoreXmlFile::openMetaAttribute()
{
	ASSERT( m_metaAttribute != NULL );
	COMTHROW( m_metaAttribute->get_AttrID(&m_metaAttributeId) );
    COMTHROW( m_metaAttribute->get_ValueType(&m_metaAttributeValType) );

	if( m_metaAttributeId == ATTRID_NONE )
		HR_THROW(E_METAPROJECT);
}

void CCoreXmlFile::closeMetaAttribute()
{
	m_metaAttribute        = NULL;
	m_metaAttributeId      = ATTRID_NONE;
    m_metaAttributeValType = VALTYPE_NONE;
}

void CCoreXmlFile::parseConnectionString( BSTR connection )
{   
    // connection string format:
    // START   = PAIRMGX | PAIR2
    // PAIR2   = PAIR PAIR2 | empty
    // PAIR    = KEY '=' VAL
    // PAIRMGX = 'MGX=' string
    // KEY     = string
    // VAL     = '"' string '"'
    //
    // valid key values: MGX, 
    //
    // example: 
    //   MGX="C:\temp\test1" vssdatabase="frfre" vsspath="dede" user="fdasfdsfds" password="defrefre"
    std::string conn;
    CopyTo(connection, conn);
    const char * connectionString = conn.c_str();

    if( strncmp( connectionString, "MGX=", 4 ) != 0 )
	    HR_THROW(E_INVALID_USAGE);
    
    int          size = conn.size();
    std::string       key;
    std::string       val;
    bool         keyCollecting = true;
    bool         startedValue  = false;

    m_folderPath     = "";
    m_vssUser        = "";
    m_vssPassword    = "";
    m_vssDatabaseStr = "";
    m_vssParentPath  = "";

    for( int i=0; i<size; ++i )
    {
        char ch = connectionString[i];
        if( keyCollecting )
        {
            if( ch == '=' )
            {
                keyCollecting = false;
                startedValue  = false;
                val = "";
            }
            else
            {
                if( ch != ' ' && ch != '\t' )
                    key += ch;
            }
        }
        else
        {
            if( startedValue )
            {
                if( ch != '\"' )
                    val += ch;
                else
                {
                    // key-val pair is finished
                    if( stricmp( key.c_str(), "MGX" ) == 0 )
                        m_folderPath = val;
                    else if( stricmp( key.c_str(), "user" ) == 0 )
                        m_vssUser = val;
                    else if( stricmp( key.c_str(), "password" ) == 0 )
                        m_vssPassword = val;
                    else if( stricmp( key.c_str(), "vssDatabase" ) == 0 )
                        m_vssDatabaseStr = val;
                    else if( stricmp( key.c_str(), "vssPath" ) == 0 )
                        m_vssParentPath = val;
                    else if( stricmp( key.c_str(), "clearCase" ) == 0 )
                        m_clearCaseString = val;

                    keyCollecting = true;
                    key = "";
                }
            }
            else
            {
                if( ch == '\"' )
                    startedValue = true;

            }
        }      
    }
}

void CCoreXmlFile::setFileNames()
{  
    char drive[_MAX_DRIVE];
    char dir[_MAX_DIR];
    char fname[_MAX_FNAME];
    char ext[_MAX_EXT];

    _splitpath( m_folderPath.c_str(), drive, dir, fname, ext );

    m_parentFolderPath = drive;
    m_parentFolderPath += dir;
    if( m_parentFolderPath[m_parentFolderPath.size()-1] == '\\' )
        m_parentFolderPath.resize(m_parentFolderPath.size()-1);

    m_projectName = fname;

    m_cacheFileName = m_folderPath;
    m_cacheFileName += "\\";
    m_cacheFileName += fname;
    m_cacheFileName += ".bin";

    m_projectFileName = m_folderPath;
    m_projectFileName += "\\";
    m_projectFileName += fname;
    m_projectFileName += ".mgx";

    m_vssPath = m_vssParentPath;
    m_vssPath += "/";
    m_vssPath += m_projectName;
}

/*void CCoreXmlFile::setFileNamesOnOpen(BSTR connection)
{
    std::string conn;

    CopyTo(connection, conn);

    if( !(std::string(conn, 0, 4) == "MGX=") )
	    HR_THROW(E_INVALID_USAGE);

    conn.erase(0, 4);
    
    char drive[_MAX_DRIVE];
    char dir[_MAX_DIR];
    char fname[_MAX_FNAME];
    char ext[_MAX_EXT];

    _splitpath( conn.c_str(), drive, dir, fname, ext );

    m_folderPath = drive;
    m_folderPath += "\\";
    m_folderPath += dir;
    m_folderPath.resize(m_folderPath.size()-1); // remove the ending "\"

    m_cacheFileName = m_folderPath;
    m_cacheFileName += "\\";
    m_cacheFileName += fname;
    m_cacheFileName += ".bin";

    m_projectFileName = m_folderPath;
    m_projectFileName += "\\";
    m_projectFileName += fname;
    m_projectFileName += ".mgx";
}*/

void CCoreXmlFile::getContainerFileName(XmlObject * obj, std::string& str, bool fullpath)
{
    ASSERT( obj->isContainer() );

    std::string guidStr;
    guid2str( obj->m_guid, guidStr );

    if( fullpath )
    {
        str = m_folderPath;
        str += "\\";
    }
    else
        str = "";

    str += guidStr;
    str += ".xml";
}

void CCoreXmlFile::getContainerName(XmlObject * obj, string& name, string& type)
{
    CComObjPtr<ICoreMetaObject>     metaobject;
    CComBSTR                        metaToken;

    COMTHROW( m_metaProject->get_Object( obj->m_metaid, PutOut(metaobject) ) );
    COMTHROW( metaobject->get_Token( &metaToken ) );

    CopyTo(metaToken, type);

    name = "";
    AttribMapIter it;
    for( it=obj->m_attributes.begin(); it!=obj->m_attributes.end(); ++it )
    {
        XmlAttrBase                    * attr = it->second;
        CComObjPtr<ICoreMetaAttribute>   metaAttrib;
        CComBSTR                         attribToken;
        string                           attribToken2;
        string                           attrVal;

        COMTHROW( metaobject->get_Attribute( it->first, PutOut(metaAttrib) ) );
        metaAttrib->get_Token( &attribToken );

        CopyTo(attribToken, attribToken2);

        if( stricmp( attribToken2.c_str(), "name" ) == 0 && attr->getType() == VALTYPE_STRING )
        {
            XmlAttrBase * attr = it->second;
            attr->toString(name);
            break;
        }
    }
}

void CCoreXmlFile::clearAll()
{
    for( XmlObjVecIter i=m_objects.begin(); i!=m_objects.end(); ++i )       
        delete (*i);
    m_objects.clear();
    m_objectsByGUID.clear();
    m_openedObject = NULL;
    m_root = NULL;
}

void CCoreXmlFile::addObject(XmlObject * obj)
{
    obj->m_index = m_objects.size();
    m_objects.push_back( obj );
    m_objectsByGUID.insert( GUIDToXmlObjectMap::value_type( obj->m_guid, obj ) );
}
    
void CCoreXmlFile::deleteObject(XmlObject * obj)
{
    // remove pointers
    for(AttribMapIter it=obj->m_attributes.begin(); it!=obj->m_attributes.end(); ++it)
    {
        if( it->second->getType() == VALTYPE_POINTER )
        {
            // remove it from parent object's collection
            XmlAttrPointer * pointer = (XmlAttrPointer*)it->second;
            if( pointer->m_parent != NULL )
            {
                AttribMapIter it2 = pointer->m_parent->m_attributes.find( it->first + ATTRID_COLLECTION );
                ASSERT( it2 != pointer->m_parent->m_attributes.end() );
                ASSERT( it2->second->getType() == VALTYPE_COLLECTION );
                ((XmlAttrCollection*)it2->second)->m_children.erase(obj);
            }
        }
        else if( it->second->getType() == VALTYPE_COLLECTION )
        {
            XmlAttrCollection * collection = (XmlAttrCollection*)it->second;
            for( XmlObjSetIter it2=collection->m_children.begin(); it2!=collection->m_children.end(); ++it2 )
            {
                // set parent of this child to NULL
                XmlObject * obj2 = (*it2);
                AttribMapIter it3 = obj2->m_attributes.find( it->first - ATTRID_COLLECTION );
                ASSERT( it3 != obj2->m_attributes.end() );
                ASSERT( it3->second->getType() == VALTYPE_POINTER );
                ((XmlAttrPointer *)(it3->second))->m_parent = NULL;
            }
        }
    }

    obj->m_deleted = true;
}

void CCoreXmlFile::setPointer(XmlObject * obj, attrid_type attribId, XmlObject * parent)
{
    ASSERT( obj!=NULL );

    AttribMapIter it = obj->m_attributes.find(attribId);
    ASSERT( it!=obj->m_attributes.end() );
    
    XmlAttrPointer * attr = (XmlAttrPointer*)it->second;
    
    // remove item from old parent's list
    if( attr->m_parent != NULL )
    {
        // find collection attribute of parent
        AttribMapIter it2 = attr->m_parent->m_attributes.find( attribId + ATTRID_COLLECTION );
        ASSERT( it2 != attr->m_parent->m_attributes.end() );
        ASSERT( it2->second->getType() == VALTYPE_COLLECTION );
        
        // remove this form the list
        ((XmlAttrCollection *)it2->second)->m_children.erase(obj);
    }

    // set pointer attribute
    attr->m_parent = parent;

    // add item to new parent's list
    if( parent != NULL )
    {
        AttribMapIter it3 = parent->m_attributes.find( attribId + ATTRID_COLLECTION );
        ASSERT( it3 != parent->m_attributes.end() );
        ASSERT( it3->second->getType() == VALTYPE_COLLECTION );
        ((XmlAttrCollection *)it3->second)->m_children.insert(obj);
    }
}

/*void CCoreXmlFile::setPointer(XmlObject * obj, XmlAttrPointer * attr, VARIANT p)
{
    metaobjidpair_type idpair;
    CopyTo(p, idpair);
    XmlObject * parent = objectFromObjId(idpair);
    setPointer( m_openedObject, m_metaAttributeId, parent );
}*/

void CCoreXmlFile::setPointer(VARIANT p)
{   
    metaobjidpair_type idpair;
    CopyTo(p, idpair);
    XmlObject * parent = objectFromObjId(idpair);
    setPointer( m_openedObject, m_metaAttributeId, parent );
}
    
void CCoreXmlFile::updateCollections()
{
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);
        for( AttribMapIter j=obj->m_attributes.begin(); j!=obj->m_attributes.end(); ++j)
        {
            if( j->second->getType() == VALTYPE_POINTER )
            {
                XmlAttrPointer * pointer = (XmlAttrPointer*)j->second;
                if( pointer->m_parent != NULL )
                {
                    AttribMapIter k = pointer->m_parent->m_attributes.find( j->first + ATTRID_COLLECTION );
                    ASSERT( k != pointer->m_parent->m_attributes.end() );
                    ((XmlAttrCollection*)k->second)->m_children.insert( obj );
                }
            }
        }
    }
}

void CCoreXmlFile::resolvePointers(UnresolvedPointerVec& pointers)
{
    for( UnresolvedPointerVecIt it=pointers.begin(); it!=pointers.end(); ++it )
    {
        if( it->m_pointedObjGuid == GUID_NULL )
        {
            setPointer( it->m_object, it->m_attrib, NULL );
        }
        else
        {
            GUIDToXmlObjectMapIter it2 = m_objectsByGUID.find( it->m_pointedObjGuid );
            if( it2 != m_objectsByGUID.end() )
            {
                setPointer( it->m_object, it->m_attrib, it2->second );
            }
            else
            {
                // TODO: invalid pointer, what to do?
            }
        }
    }
}

void CCoreXmlFile::createSourceControlInfoRegNode(XmlObject * obj)
{
    ASSERT( obj->m_metaid==DTID_MODEL || obj->m_metaid==DTID_FOLDER );

    CComObjPtr<ICoreMetaObject> regNodeMetaObj;
	COMTHROW( m_metaProject->get_Object(DTID_REGNODE, PutOut(regNodeMetaObj)) );
	ASSERT( regNodeMetaObj != NULL );

    // create and add new object
    XmlObject * obj2 = new XmlObject(regNodeMetaObj,true);
    obj2->m_sourceControlInfoRegNode = true;
    addObject( obj2 );

    // set parent
    setPointer( obj2, ATTRID_REGNOWNER, obj );

    // set attributes
    AttribMapIter it = obj2->m_attributes.find( ATTRID_NAME );
    it->second->fromString("SourceControlInfo");
    it = obj2->m_attributes.find( ATTRID_REGNODEVALUE );
    it->second->fromString("0");
    it = obj2->m_attributes.find( ATTRID_REGFLAGS );
    it->second->fromString("1");
}

void CCoreXmlFile::createSourceControlInfoRegNodes()
{
    // collect containers
    XmlObjVec containers;
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);

        if( obj->m_metaid==DTID_MODEL || obj->m_metaid==DTID_FOLDER )
            containers.push_back( obj );
        //if( obj->isContainer() )
          //  containers.push_back( obj );
    }

    for( it=containers.begin(); it!=containers.end(); ++it )
        createSourceControlInfoRegNode( *it );
}

void CCoreXmlFile::getPointer(XmlAttrPointer * attr, VARIANT * p)
{
    ASSERT( attr!= NULL );
    metaobjidpair_type id;
    objIdFromObject( attr->m_parent, id );
    CopyTo(id, p);
}

void CCoreXmlFile::getCollection(XmlAttrCollection * attr, VARIANT * p)
{
    ASSERT( attr!= NULL );
    std::vector<metaobjidpair_type> idpairs;
    for( XmlObjSetIter it = attr->m_children.begin(); it != attr->m_children.end(); ++it )
    {
        metaobjidpair_type id;
        objIdFromObject( *it, id );
        idpairs.push_back( id );
    }
    CopyTo(idpairs, p);
}

// TODO: test!
XmlObject * CCoreXmlFile::getContainer(XmlObject * obj)
{
    ASSERT( obj != NULL );
    XmlObject * container = obj;
    while( container!=NULL && !container->isContainer() )
    {
        ParentMap::iterator it = m_parentMap.find( container->m_metaid );
        ASSERT( it != m_parentMap.end() );
        AttribMapIter it2 = container->m_attributes.find( it->second );
        ASSERT( it2 != container->m_attributes.end() );
        container = ((XmlAttrPointer*)(it2->second))->m_parent;    
    }
    return container;
}

// TODO: test!
void CCoreXmlFile::getContainedObjects(XmlObject * obj, XmlObjVec& vec)
{
    ASSERT( obj != NULL );
    vec.push_back(obj);

    for(AttribMapIter it=obj->m_attributes.begin(); it!=obj->m_attributes.end(); ++it)
    {
        if( it->second->getType() == VALTYPE_COLLECTION )
        {
            XmlAttrCollection * coll = (XmlAttrCollection *)it->second;
            for( XmlObjSetIter it2=coll->m_children.begin(); it2!=coll->m_children.end(); ++it2 )
            {
                XmlObject * obj2 = (*it2);
                if( obj2->m_metaid != DTID_MODEL && obj2->m_metaid != DTID_FOLDER )
                {
                    ParentMap::iterator it3 = m_parentMap.find( obj2->m_metaid );
                    ASSERT( it3 != m_parentMap.end() );
                    if( it3->first + ATTRID_COLLECTION == it->first )
                        getContainedObjects(obj2,vec);
                }
            }
        }
    }
}

void CCoreXmlFile::getCheckOutContainers(XmlObjSet& objects, XmlObjSet& containers)
{
    XmlObjSet processedObjects;
    for( XmlObjSetIter it=objects.begin(); it!=objects.end(); ++it )
        getDependetContainers( *it, containers, processedObjects );
}

void CCoreXmlFile::getDependetContainers(XmlObject * obj, XmlObjSet& containers, XmlObjSet& processedObjects)
{
    if( obj!= NULL && processedObjects.find(obj)==processedObjects.end() )  // not processed yet
    {
        processedObjects.insert(obj);
        XmlObject * container = getContainer(obj);
        if( container != NULL )
        {
            //TODO: ezt nagyon at kell meg gondolni. Ha pl torles van kerdes:
            // allitjak-e az attributumokat olyan sorrandben hogy nem minden container kerul
            // be a listaba (ahogy a kapcsolatok megszunnek elvesztjuk a lehetsoeget hogy a 
            // a contaniert megtlaljuk)
            // lehet hogy a containert megiscsak tarolni kene??? ezt at kell gondolni!!!
            containers.insert( container );
            for(AttribMapIter it=container->m_attributes.begin(); it!=container->m_attributes.end(); ++it)
            {
                if( it->second->getType() == VALTYPE_COLLECTION )
                {
                    XmlAttrCollection * coll = (XmlAttrCollection*)it->second;
                    for( XmlObjSetIter it2=coll->m_children.begin(); it2!=coll->m_children.end(); ++it2 )
                    {
                           XmlObject * obj2 = (*it2);
                        if( obj2 != NULL )
                            getDependetContainers( obj2, containers, processedObjects );
                    }
                }
            }
        }
    }
}

/* 
Ask user if files need to be checked out.

Returns true if all the files has been checked out successfully.
*/
bool CCoreXmlFile::checkOutFiles(XmlObjSet& containers)
{
    AFX_MANAGE_STATE(AfxGetStaticModuleState( ));

    if( containers.size() == 0 )
        return true;

    XmlObjSetIter it;
    XmlObjSet     containersUsedByOthers;
    
    // count the files need to be checked out
    XmlObjSet readOnlyFiles;
    bool      checkdOutByOthers = false;
    for( it=containers.begin(); it!=containers.end(); ++it )
    {
        XmlObject * obj = *it;
        ASSERT( obj != NULL );

        if( m_sourceControl == SC_CLEARCASE )
        {
            string fileName;
            getContainerFileName( obj, fileName, true );

            FILE * f = fopen( fileName.c_str(), "r" );
            if( f!=NULL )
            {
                fclose(f);
                int checkoutState = getCheckOutStateCC( fileName.c_str() );

                if( checkoutState == CS_OTHER_USER )
                {
                    checkdOutByOthers = true;
                    readOnlyFiles.insert( obj );
                    containersUsedByOthers.insert( obj );
                }
                else if( checkoutState == SC_NONE )
                    readOnlyFiles.insert( obj );
            }
        }
        else if( isContainerReadOnly(obj) )
        {
            readOnlyFiles.insert( obj );
            if( m_sourceControl != SC_NONE )
            {
                // check if checked out by other users
                if( isContinerCheckedOut( obj ) )
                {
                    containersUsedByOthers.insert( obj );
                    checkdOutByOthers = true;
                }
            }
        }
    }

    if( readOnlyFiles.size() == 0 )
        return true;

    // if there is no source control nothing we can do, notify the user and roll back transaction
    if( m_sourceControl == SC_NONE )
    {
        AfxMessageBox( "Cannot perform this operation because some read only files need modification. Maybe the project is under source control. Try to open it again end login to the source control database!" );
        return false;
    }

    // if there are files checkd out by other we are done, roll back transaction
    if( checkdOutByOthers )
    {
        CFilesInUseDlg dlg;
        if( dlg.DoModal() == IDOK )
            showUsedFiles( containersUsedByOthers );
        return false;
    }

    // ask user for check out confirmation
    if( AfxMessageBox( "To perform this operation some files will be checked out. Do you want to continue?", MB_YESNO ) == IDYES )
    {
        getLatestVersion();

        bool needClose = filesModifedByOthers();

        if( needClose )
        {
            AfxMessageBox( "Part of the project has been modified by other users, it is highly recomended to close and reopen your project to synchronize it!");
            return false;
        }

        // check out files
        try
        {
            for( it=readOnlyFiles.begin(); it!=readOnlyFiles.end(); ++it )
                checkOutContainer( *it );
            
        }
        catch(...)
        {
            // TODO: megse sikerult kicsekkelni mindent, hogyan tovabb?
            // vissza kell undocheckout-olni amit sikerult es roll back, uzenet a userbek hogy 
            // gaz van.
            return false;
        }
                               

        return true;
    }
    else
        return false;
}

XmlObject * CCoreXmlFile::objectFromObjId(metaobjidpair_type idpair)
{
    if( idpair.metaid == METAID_NONE && idpair.objid == OBJID_NONE )
        return NULL;

    if( idpair.metaid == METAID_ROOT )
        return m_root;
    else
        return (XmlObject*)idpair.objid;
}

void CCoreXmlFile::objIdFromObject(XmlObject * obj, metaobjidpair_type& idpair)
{
    if( obj == NULL )
    {
        idpair.metaid = METAID_NONE;
        idpair.objid  = OBJID_NONE;
    }
    else
    {
        idpair.metaid = obj->m_metaid;
        if( idpair.metaid == METAID_ROOT )
            idpair.objid = 1;
        else
            idpair.objid = (long)obj;
    }
}

void CCoreXmlFile::writeBinaryCache()
{
    FILE * f = fopen( m_cacheFileName.c_str(), "wb" );
    if( f==NULL )
        HR_THROW(E_FILEOPEN);

    XmlObjVecIter i;
    metaid_type   mid;

    // write out GUIDs and metaids
    int n = m_objects.size();
    fwrite( &n, sizeof(n), 1, f );
    for( i = m_objects.begin(); i != m_objects.end(); ++i )
    {
        mid = (*i)->m_metaid;
        fwrite( &mid, sizeof(mid), 1, f );
        fwrite( &((*i)->m_guid), sizeof(GUID), 1, f );
    }

    // write out pointers
    int x;
    for( i=m_objects.begin(); i!=m_objects.end(); ++i )
    {   
        XmlObject * obj = (*i);
        for( AttribMapIter j=obj->m_attributes.begin(); j!=obj->m_attributes.end(); ++j)
        {
            if( j->second->getType() == VALTYPE_POINTER )
            {
                XmlAttrPointer * pointer = (XmlAttrPointer*)j->second;
                if( pointer->m_parent == NULL )
                    x = -1;
                else
                    x = pointer->m_parent->m_index;
                fwrite( &x, sizeof(x), 1, f );
            }
        }
    }

    fclose(f);
}

bool CCoreXmlFile::readBinaryCache()
{    
    FILE * f = fopen( m_cacheFileName.c_str(), "rb" );
    if( f==NULL )
        return false;

    clearAll();

    // read GUIDs and metaids and create objects
    int n,i;
    fread( &n, sizeof(n), 1, f );
    for( i=0; i<n; ++i )
    {
        metaid_type  metaid;
        GUID         guid;

        fread( &metaid, sizeof(metaid), 1, f );
        fread( &guid, sizeof(GUID), 1, f );

        CComObjPtr<ICoreMetaObject> metaobject;
        COMTHROW( m_metaProject->get_Object(metaid, PutOut(metaobject)) );

        XmlObject * obj = new XmlObject(metaobject,false);
        obj->m_guid   = guid;
        addObject(obj);

        if( metaid == METAID_ROOT )
            m_root = obj;
    }

    UnresolvedPointerVec pointers;
    int x;
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);
        for( AttribMapIter j=obj->m_attributes.begin(); j!=obj->m_attributes.end(); ++j)
        {
            if( j->second->getType() == VALTYPE_POINTER )
            {
                XmlAttrPointer * pointer = (XmlAttrPointer*)j->second;

                UnresolvedPointer p;
                p.m_object = obj;
                p.m_attrib = j->first;
                                
                fread( &x, sizeof(x), 1, f );
                if( x == -1 )
                    p.m_pointedObjGuid = GUID_NULL;
                else
                    p.m_pointedObjGuid = m_objects[x]->m_guid;

                pointers.push_back( p );
            }
        }
    }

    resolvePointers( pointers );

    // read and set pointers
    /*int x;
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);
        for( AttribMapIter j=obj->m_attributes.begin(); j!=obj->m_attributes.end(); ++j)
        {
            if( j->second->getType() == VALTYPE_POINTER )
            {
                XmlAttrPointer * pointer = (XmlAttrPointer*)j->second;

                fread( &x, sizeof(x), 1, f );
                if( x == 0 )
                    pointer->m_parent = NULL;
                else
                    pointer->m_parent = m_objects[x];
            }
        }
    }*/



    fclose(f);

    //updateCollections();

    return true;
}

void CCoreXmlFile::createProjectFile()
{    
    // in case of clearcase directory must be created with clearcase (no addelement, grrrr!)
    if( m_sourceControl == SC_CLEARCASE )
    {
        try
        {
            // create and add project folder to source control
            checkOutFileCC( m_parentFolderPath.c_str() );   // maybe this not needed if we keep it checked out
            addDirToCC( m_folderPath.c_str() );
        }
        catch(...)
        {
            AfxMessageBox("ClearCase error! Cannot add project to ClearCase. Errocode=1");
            m_sourceControl = SC_NONE;
        }
    }
    else
    {
        // create project folder
        if( CreateDirectory(m_folderPath.c_str(),NULL) == 0 )
            HR_THROW(E_FILEOPEN);
    }

    // create projet file
    FILE * f = fopen( m_projectFileName.c_str(), "wt" );
    if( f == NULL )
        HR_THROW(E_FILEOPEN);
    fprintf( f, "<GME " );
    if( m_vssDatabaseStr.size() > 0 )
        fprintf( f, "VSSDatabase=\"%s\" VSSPath=\"%s\"", m_vssDatabaseStr.c_str(), m_vssPath.c_str() );
    fprintf( f, ">\n" );
    fprintf( f, "</GME>\n" );
    fclose( f );

    // if project under source control, the project file should be added to source control    
    if( m_sourceControl == SC_SOURCESAFE )
    {
        // get project item
        CComObjPtr<IVSSItem> projectItem;
        CComBSTR path = m_vssPath.c_str();
        COMTHROW( m_vssDatabase->get_VSSItem( path, false, &(projectItem.p)) );

        // add to source control projecct
        CComObjPtr<IVSSItem> item;
        CComBSTR fileName2 = m_projectFileName.c_str();
        
        projectItem->Add( fileName2, NULL, 0, &(item.p));   // TODO error handling
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        try
        {
            // add project file
            addFileToCC( m_projectFileName.c_str() );
            //checkInFileCC( m_projectFileName.c_str() );
            //checkInFileCC( m_folderPath.c_str() );
            //checkInFileCC( m_parentFolderPath.c_str() );
        }
        catch(...)
        {
            AfxMessageBox("ClearCase error! Cannot add project to ClearCase. Errocode=2");
            m_sourceControl = SC_NONE;
        }
    }
}

void CCoreXmlFile::readProjectFile()
{
    // Project file is an xml file, with a GME tag.
    // possible attributes: VSSDatabase, VSSPath
    // example: <GME VSSDatabase="\\bogyom\GMEXMLBackEndTest\sorucesafedb\srcsafe.ini" VSSPath="$\test1"></GME>


    DOMBuilder * parser = NULL;
    DOMImplementationLS * domimpl = DOMImplementationRegistry::getDOMImplementation(NULL);
    ASSERT( domimpl != NULL );
    parser = domimpl->createDOMBuilder( DOMImplementationLS::MODE_SYNCHRONOUS, NULL );
    ASSERT( parser != NULL );

    DOMDocument * doc = parser->parseURI( m_projectFileName.c_str() );
    if( doc == NULL )
        HR_THROW(E_FILEOPEN);
    DOMElement * e = doc->getDocumentElement();

    char * vssDatabase = XMLString::transcode(e->getAttribute(XMLString::transcode("VSSDatabase")));
    if( strlen(vssDatabase) != 0 )
    {
        m_vssDatabaseStr = vssDatabase;
        char * vssPath   = XMLString::transcode(e->getAttribute(XMLString::transcode("VSSPath")));

        m_vssPath = vssPath;
        delete vssPath;
        openSourceSafeDatabase();
    }
    else
    {        
        m_sourceControl = SC_NONE;
        m_vssDatabaseStr = "";
        m_vssPath        = "";

        // check if the project is under clearcase
        if( createClearCaseApp() )
        {
            if( isPathUnderClearCase( m_parentFolderPath.c_str() ) )
                m_sourceControl = SC_CLEARCASE;
        }
    }
    delete vssDatabase;
    delete parser;
}

void CCoreXmlFile::writeAll()
{
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {
        XmlObject * obj = (*it);
        if( obj->isContainer() && obj->m_loaded )
            writeXMLFile( obj );

        // itt csak a modositottakat kellene kiirni nem?
    }
}

void CCoreXmlFile::writeXMLFile(XmlObject * container)
{
    if( !container->isContainer() )
        HR_THROW(E_INVALID_USAGE);

    std::string fileName;
    getContainerFileName(container, fileName);

    time_t        currentTime1 = time( &currentTime1 );
    CTime         currentTime2( currentTime1 + 1 );

    // open file
    FILE * file = fopen( fileName.c_str(), "w" );
    if( file != NULL )
    {
        // write objects recursively
        writeObject( container, file, true, "", currentTime2 );
    
        fclose(file);

        // add to source control if not added yet
        try
        {
            addToSourceControl( container );
        }
        catch(...)
        {
            // TODO:
        }
    }
    else
    {
        // TODO: the file is probably read only (it means there was no change)
    }
}

void CCoreXmlFile::writeObject(XmlObject * obj, FILE * file, bool container, const char * prefix, CTime lastWriteTime )
{        
    std::string                     str;
    CComObjPtr<ICoreMetaObject>     metaobject;
    CComBSTR                        metaToken;

    if( obj->m_sourceControlInfoRegNode )
        return;

    obj->m_lastWriteTime = lastWriteTime;

    COMTHROW( m_metaProject->get_Object( obj->m_metaid, PutOut(metaobject) ) );
    COMTHROW( metaobject->get_Token( &metaToken ) );
    guid2str( obj->m_guid, str );
    
    fprintf(file, "%s<%S MetaId=\"%d\" Id=\"%s\"", prefix, metaToken, obj->m_metaid, str.c_str() );
    if( obj->m_deleted )
        fprintf(file, " deleted=\"true\"");

    // write pointers, attributes
    AttribMapIter it;
    for( it=obj->m_attributes.begin(); it!=obj->m_attributes.end(); ++it )
    {
        XmlAttrBase                    * attr = it->second;
        CComObjPtr<ICoreMetaAttribute>   metaAttrib;
        CComBSTR                         attribToken;
        std::string                           attrVal;

        COMTHROW( metaobject->get_Attribute( it->first, PutOut(metaAttrib) ) );
        metaAttrib->get_Token( &attribToken );

        if( attr->getType() == VALTYPE_POINTER )
        {
            ParentMap::iterator it2 = m_parentMap.find( obj->m_metaid );
            if( it2==m_parentMap.end() || it2->second!=it->first || container )
            {
                XmlAttrPointer * pointer = (XmlAttrPointer*)attr;
                if( pointer->m_parent != NULL )
                    guid2str( pointer->m_parent->m_guid, attrVal );
            }
        }
        else if( attr->getType() != VALTYPE_COLLECTION && attr->getType() != VALTYPE_LOCK )
        {
            XmlAttrBase * attr = it->second;
            attr->toString(attrVal);
        }
                        
        if( !attrVal.empty() )
        {
            // replace spaces in attribToken with undescores
            char buf[200];
            sprintf(buf, "%S", attribToken );        
            std::string attribToken2 = buf;
            replaceSpaceWithUnderscore(attribToken2);

            string encodedAttrVal;
            encodeXMLString( attrVal, encodedAttrVal );

            fprintf(file, " %s=\"%s\"", attribToken2.c_str(), encodedAttrVal.c_str() );
        }
    }
    fprintf(file, ">\n" );

    // write out children
    // child is written if it is not the root, a model or a folder
    // and the parent of the child is a us according to m_parentMap
    std::string newPrefix = prefix;
    newPrefix += "\t";
    for( it=obj->m_attributes.begin(); it!=obj->m_attributes.end(); ++it )
    {
        XmlAttrBase * attr = it->second;
        if( attr->getType() == VALTYPE_COLLECTION )
        {
            XmlAttrCollection * coll = (XmlAttrCollection*)attr;
            for( XmlObjSetIter it2=coll->m_children.begin(); it2!=coll->m_children.end(); ++it2 )
            {
                XmlObject * obj2 = (*it2);
                if( obj2!=NULL && !obj2->isContainer() )
                {                    
                    ParentMap::iterator it3 = m_parentMap.find( obj2->m_metaid );
                    ASSERT( it3 != m_parentMap.end() );
                    if( it3->second + ATTRID_COLLECTION == it->first )
                        writeObject( obj2, file, false, newPrefix.c_str(), lastWriteTime );
                }
            }
        }
    }
    
    fprintf(file, "%s</%S>\n",prefix, metaToken);
}

/*
    Moves to continer and all contained object to disk if possible (all lock values are 0)
    It deletes secondary attributes from memory.

    Returns true if the objects have been moved to disk successfully
*/
bool CCoreXmlFile::moveContainerToDisk(XmlObject * container)
{
    XmlObjVec     objs;
    XmlObjVecIter it;

    getContainedObjects( container, objs );
    for( it=objs.begin(); it!=objs.end(); ++it )
    {
        XmlObject * obj = *it;

        AttribMapIter it2 = obj->m_attributes.find(ATTRID_LOCK);
        ASSERT( it2 != obj->m_attributes.end() );

        XmlAttrLock * lock = (XmlAttrLock*)(it2->second);
        if( lock != 0 )
            return false;
    }

    // write all objects to disk
    writeXMLFile( container );

    // unload secondary attributes from memory
    for( it=objs.begin(); it!=objs.end(); ++it )
    {
        XmlObject * obj = *it;
        obj->deleteSecondaryAttribs();
    }
    
    return false;
}

void CCoreXmlFile::fullReadContainer(XmlObject * container)
{
    std::string fileName;
    getContainerFileName(container, fileName);

    UnresolvedPointerVec pointers;
    readXMLFile( fileName.c_str(), pointers, true );
}

void CCoreXmlFile::readXMLFile( const char * fileName, UnresolvedPointerVec& pointers, bool fullLoad )
{
    DOMBuilder * parser = NULL;
    try
    {    
        DOMImplementationLS * domimpl = DOMImplementationRegistry::getDOMImplementation(NULL);
        ASSERT( domimpl != NULL );

        parser = domimpl->createDOMBuilder( DOMImplementationLS::MODE_SYNCHRONOUS, NULL );
        ASSERT( parser != NULL );

        // get last write time
        WIN32_FILE_ATTRIBUTE_DATA attr;        
        GetFileAttributesEx( fileName, GetFileExInfoStandard, &attr );
        CTime lastWriteTime( attr.ftLastWriteTime );

        DOMDocument * doc = parser->parseURI( fileName );  
        readObject( doc->getDocumentElement(), pointers, NULL, fullLoad, lastWriteTime );

        delete parser;
    }
    catch(...)
    {
        if( parser != NULL )
            delete parser;

        HR_THROW(E_FILEOPEN);
    }
}

void CCoreXmlFile::readObject(DOMElement * e, UnresolvedPointerVec& pointers, XmlObject * parent, bool fullLoad, CTime lastWriteTime )
{   
    // if the object is deleted do not deal with it
    char * deletedStr = XMLString::transcode(e->getAttribute(XMLString::transcode("deleted")));
    bool deleted = (strcmp( deletedStr, "true" ) == 0);
    delete deletedStr;
    if( deleted )
        return;
   
    // get metaid, and id
    char * metaidStr  = XMLString::transcode(e->getAttribute(XMLString::transcode("MetaId")));
    char * objGUIDStr = XMLString::transcode(e->getAttribute(XMLString::transcode("Id" )));
    long   metaid     = atoi( metaidStr );
    GUID   guid       = str2guid( objGUIDStr );
    delete metaidStr;
    delete objGUIDStr;

    // get meta object
    CComObjPtr<ICoreMetaObject> metaobject;
#pragma warning( disable: 4244) // conversion from 'long' to 'short', possible loss of data
    COMTHROW( m_metaProject->get_Object( metaid, PutOut(metaobject) ) );
#pragma warning( default: 4244) // conversion from 'long' to 'short', possible loss of data

    // find or create object
    XmlObject * obj = NULL;
    GUIDToXmlObjectMapIter it = m_objectsByGUID.find( guid );
    if( it != m_objectsByGUID.end() )
    {
        obj = it->second;
        if( !obj->m_loaded && fullLoad )
        {
            obj->createAttributes(metaobject,XmlObject::ATTR_SECONDARY);
            obj->m_loaded = true;
        }
    }
    else
    {
        obj = new XmlObject(metaobject,fullLoad);
        obj->m_guid = guid;
        addObject( obj );
        if( metaid == METAID_ROOT )
            m_root = obj;
    }

    obj->m_deleted = false;
    obj->m_lastWriteTime = lastWriteTime;
    
    // read attributes
    AttribMapIter it2;
    for( it2 = obj->m_attributes.begin(); it2 != obj->m_attributes.end(); ++it2 )
    {        
        CComObjPtr<ICoreMetaAttribute>  metaAttrib;
        CComBSTR                        attribToken;

        COMTHROW( metaobject->get_Attribute( it2->first, PutOut(metaAttrib) ) );
        COMTHROW( metaAttrib->get_Token( &attribToken ));        

        char * attrVal = XMLString::transcode(e->getAttribute(attribToken));

        replaceUnderscoreWithSpace( attrVal );

        XmlAttrBase * attr = it2->second;
        if( attr->getType() == VALTYPE_POINTER )
        {
            UnresolvedPointer p;
            p.m_object = obj;
            p.m_attrib = it2->first;          
            if( attrVal==NULL || strlen(attrVal)==0 )
                p.m_pointedObjGuid = GUID_NULL;
            else
                p.m_pointedObjGuid = str2guid( attrVal );
            pointers.push_back( p );
        }
        else if( attr->getType() != VALTYPE_LOCK && attr->getType() != VALTYPE_COLLECTION )
        {
            it2->second->fromString(attrVal);
        }

        delete attrVal;
    }

    // implicit parent pointer
    if( parent != NULL )
    {
        ParentMap::iterator it3 = m_parentMap.find( obj->m_metaid );
        UnresolvedPointer p;
        p.m_object = obj;
        p.m_attrib = it3->second;
        p.m_pointedObjGuid = parent->m_guid;
        pointers.push_back(p);
    }        
  
    // read children
    DOMNodeList * children = e->getChildNodes();
    for( int i=0; i< (int) children->getLength(); ++i )
    {
        DOMNode * node = children->item(i);
        if( node->getNodeType() == DOMNode::ELEMENT_NODE )        
            readObject( (DOMElement*)node, pointers, obj, fullLoad, lastWriteTime );
    }
}

void CCoreXmlFile::readAll( bool fullLoad )
{
    UnresolvedPointerVec pointers;
    char                 buf[_MAX_PATH];
    _finddata_t          fileInfo;

    // todo: preconditions

    //clearAll();

    sprintf( buf, "%s\\*.xml", m_folderPath.c_str() );
    long searchHandle = _findfirst( buf, &fileInfo );
    long ret = searchHandle;
    while( ret != -1 )
    {
        sprintf( buf, "%s\\%s", m_folderPath.c_str(), fileInfo.name );
        readXMLFile( buf, pointers, fullLoad );
        ret = _findnext( searchHandle, &fileInfo );
    }
    _findclose( searchHandle );

    
    resolvePointers( pointers );

    
    createSourceControlInfoRegNodes();
    updateSourceControlInfo();

}

void CCoreXmlFile::getLatestAndLoad()
{
    char          buf[_MAX_PATH];
    _finddata_t   fileInfo;
    time_t        currentTime1 = time( &currentTime1 );
    CTime         currentTime2( currentTime1 - 100 );

    // get latest version from source control
    getLatestVersion();

    // itarete on files and read new and modified files
    sprintf( buf, "%s\\*.xml", m_folderPath.c_str() );
    long searchHandle = _findfirst( buf, &fileInfo );
    long ret = searchHandle;
    while( ret != -1 )
    {
        sprintf( buf, "%s\\%s", m_folderPath.c_str(), fileInfo.name );

        WIN32_FILE_ATTRIBUTE_DATA attr;        
        if( GetFileAttributesEx( buf, GetFileExInfoStandard, &attr ) )
        {
            CTime modTime( attr.ftLastWriteTime );
            if( modTime >= currentTime2 )
            {
                // this is a new or modified file, read it
                UnresolvedPointerVec pointers;
                readXMLFile( buf, pointers, false );
            }
        }
        
        ret = _findnext( searchHandle, &fileInfo );
    }
    _findclose( searchHandle );
}

bool CCoreXmlFile::getSourceSafeUserInfo()
{
    CVSSLoginDlg dlg;
    dlg.m_project  = m_projectFileName.c_str();
    dlg.m_database = m_vssDatabaseStr.c_str();
    dlg.m_user     = "";
    dlg.m_password = "";

    if( dlg.DoModal() == IDOK )
    {
        m_vssUser     = dlg.m_user;
        m_vssPassword = dlg.m_password;
        return true;
    }

    return false;
}

void CCoreXmlFile::createSourceSafeDatabase()
{
    AFX_MANAGE_STATE(AfxGetStaticModuleState( ));

    m_sourceControl = SC_NONE;

    // create source control project if specified one
    if( m_vssDatabaseStr.size() > 0 )
    {
        if( m_vssUser.size() > 0 || getSourceSafeUserInfo() )
        {
            // try to open the sorusafe database
            try
            {
                COMTHROW( CoCreateInstance(CLSID_VSSDatabase, NULL, CLSCTX_ALL, IID_IVSSDatabase, (void**)&(m_vssDatabase.p) ));

                CComBSTR path  = m_vssDatabaseStr.c_str();
                CComBSTR user2 = m_vssUser.c_str();
                CComBSTR pwd2  = m_vssPassword.c_str();

                COMTHROW( m_vssDatabase->Open( path, user2, pwd2 ) );

                CComObjPtr<IVSSItem> item;
                CComBSTR parentProject = m_vssParentPath.c_str();
                COMTHROW( m_vssDatabase->get_VSSItem( parentProject, false, &(item.p)) );

                CComObjPtr<IVSSItem> item2;
                CComBSTR subProjectName = m_projectName.c_str();
                COMTHROW( item->NewSubproject( subProjectName, NULL, &(item2.p) ));
                
                CComBSTR localSpec = m_folderPath.c_str();
                COMTHROW( item2->put_LocalSpec(localSpec) );

                m_sourceControl = SC_SOURCESAFE;
            }
            catch(...)
            {
                // error during the sourcesafe project creation

            }
        }
        else
        {
            // not logged into sourcesafe
        }
    }

    if( m_sourceControl == SC_NONE )
    {
        // TODO: error message
    }

}

void CCoreXmlFile::openSourceSafeDatabase()
{
    AFX_MANAGE_STATE(AfxGetStaticModuleState( ));

    if( m_vssUser.size() > 0 || getSourceSafeUserInfo() )
    {
        // try to open the sorusafe database
        COMTHROW( CoCreateInstance(CLSID_VSSDatabase, NULL, CLSCTX_ALL, IID_IVSSDatabase, (void**)&(m_vssDatabase.p) ));

        CComBSTR path  = m_vssDatabaseStr.c_str();
        CComBSTR user2 = m_vssUser.c_str();
        CComBSTR pwd2  = m_vssPassword.c_str();

        HRESULT hr = m_vssDatabase->Open( path, user2, pwd2 );

        if( hr == 0 )
        {
            m_sourceControl = SC_SOURCESAFE;

            CComObjPtr<IVSSItem> item;
            CComBSTR path2 = m_vssPath.c_str();
            COMTHROW( m_vssDatabase->get_VSSItem( path2, false, &(item.p)) );

            CComBSTR localSpec = m_folderPath.c_str();
            COMTHROW( item->put_LocalSpec(localSpec) );
        }
        else
        {
            char buf[200];
            sprintf( buf, "Could not connet to SurceSafe database. Error code: %x (%d). You may lose synchronization with source control. You won't be able to modify read only files.", hr, hr );
            AfxMessageBox( buf, MB_ICONEXCLAMATION );
            m_sourceControl = SC_NONE;
        }
    }
    else
    {
        AfxMessageBox( "You did not logged in the SourSafe database. You may lose synchronization with source control. You won't be able to modify read only files.", MB_ICONEXCLAMATION );
        m_sourceControl = SC_NONE;
    }
}

void CCoreXmlFile::getSourceSafePath(XmlObject * obj, std::string& str)
{
    ASSERT( m_sourceControl == SC_SOURCESAFE );

    std::string fileName;
    getContainerFileName( obj, fileName, false );

    str = m_vssPath;
    str += "\\";
    str += fileName;
}

bool CCoreXmlFile::isContainerReadOnly(XmlObject * obj)
{
    ASSERT( obj != NULL );

    std::string fileName;
    getContainerFileName( obj, fileName );

    WIN32_FILE_ATTRIBUTE_DATA attr;

    if( GetFileAttributesEx( fileName.c_str(), GetFileExInfoStandard, &attr ) )
    {
        return attr.dwFileAttributes & FILE_ATTRIBUTE_READONLY;
    }
    else
    {
        // the file does not exiest (not written out yet)
        return false;
    }
}

bool CCoreXmlFile::isContinerCheckedOut(XmlObject * obj)
{
    ASSERT( m_sourceControl != SC_NONE );

    if( m_sourceControl == SC_SOURCESAFE )
    {
        std::string fullPath;
        getSourceSafePath( obj, fullPath );
        CComBSTR fullPath2 = fullPath.c_str();

        CComObjPtr<IVSSItem> item;        
        long checkOutState;
        COMTHROW( m_vssDatabase->get_VSSItem( fullPath2, false, &(item.p)) );
        COMTHROW( item->get_IsCheckedOut(&checkOutState) );

        return checkOutState == VSSFILE_CHECKEDOUT;
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        std::string fileName;
        getContainerFileName( obj, fileName, true );
        return isFileCheckedOutCC( fileName.c_str() );
    }

    return false;
}

void CCoreXmlFile::getSSCheckOutUser(XmlObject * obj, string& user)
{
    ASSERT( m_sourceControl == SC_SOURCESAFE );
    ASSERT( obj->isContainer() );

    char buf[300];

    string fullPath;
    getSourceSafePath( obj, fullPath );
    CComBSTR fullPath2 = fullPath.c_str();

    CComObjPtr<IVSSItem> item;
    COMTHROW( m_vssDatabase->get_VSSItem( fullPath2, false, &(item.p)) );
    CComObjPtr<IVSSCheckouts> checkouts;
    COMTHROW( item->get_Checkouts( &(checkouts.p) ));
    long checkoutNum;
    COMTHROW( checkouts->get_Count( &checkoutNum ) );
    if( checkoutNum > 0 )
    {
        CComObjPtr<IVSSCheckout> checkout;
        VARIANT index;
        index.vt = VT_INT;
        index.intVal = 1;
        COMTHROW( checkouts->get_Item(index, &(checkout.p)));

        CComBSTR userName;
        COMTHROW( checkout->get_Username( &userName ));

        string name, type;
        getContainerName( obj, name, type );
       
        sprintf( buf, "%S", userName );

        user = buf;
    }
    else
    {
        // not checked out
        user = "";
    }
}
    
void CCoreXmlFile::checkOutContainer(XmlObject * obj)
{
    ASSERT( m_sourceControl != SC_NONE );

    if( m_sourceControl == SC_SOURCESAFE )
    {
        std::string fullPath;
        getSourceSafePath( obj, fullPath );
        CComBSTR fullPath2 = fullPath.c_str();

        CComObjPtr<IVSSItem> item;
        COMTHROW( m_vssDatabase->get_VSSItem( fullPath2, false, &(item.p)) );
        COMTHROW( item->Checkout( L"", NULL, NULL ) );
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        std::string fileName;
        getContainerFileName( obj, fileName, true );
        checkOutFileCC( fileName.c_str() );
    }
}

void CCoreXmlFile::addToSourceControl(XmlObject * container)
{
    ASSERT( container->isContainer() );

    std::string fileName;
    getContainerFileName(container, fileName);

    if( m_sourceControl == SC_SOURCESAFE )
    {
        // get project item
        CComObjPtr<IVSSItem> projectItem;
        CComBSTR path = m_vssPath.c_str();
        COMTHROW( m_vssDatabase->get_VSSItem( path, false, &(projectItem.p)) );

        // add to source control projecct
        CComObjPtr<IVSSItem> item;
        CComBSTR fileName2 = fileName.c_str();
        if( projectItem->Add( fileName2, NULL, 0, &(item.p)) == 0 )
        {
            // check it out
            checkOutContainer( container );
        }
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        // parent folder must be checked out
        addFileToCC( fileName.c_str() );
    }
}

void CCoreXmlFile::getLatestVersion()
{
    if( m_sourceControl == SC_SOURCESAFE )
    {
        CComObjPtr<IVSSItem> projectItem;
        CComBSTR path = m_vssPath.c_str();
        COMTHROW( m_vssDatabase->get_VSSItem( path, false, &(projectItem.p)) );
        COMTHROW( projectItem->Get( NULL, 0 ) );
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        // if we use dynamic views we do not need it
        //getLatestVerCC( m_folderPath.c_str() );
    }
}

void CCoreXmlFile::checkInAll()
{
    AFX_MANAGE_STATE(AfxGetStaticModuleState( ));

    if( AfxMessageBox( "Keep files checked out?", MB_YESNO ) == IDYES )
        checkInAll( true );
    else
        checkInAll( false );
}

void CCoreXmlFile::checkInAll( bool keepCheckedOut )
{
    if( m_sourceControl == SC_SOURCESAFE )
    {
        // source safe
        CComObjPtr<IVSSItem> projectItem;
        CComBSTR path = m_vssPath.c_str();
        COMTHROW( m_vssDatabase->get_VSSItem( path, false, &(projectItem.p)) );
        if( keepCheckedOut )
            COMTHROW( projectItem->Checkin( NULL, NULL, VSSFLAG_KEEPYES ) );
        else
            COMTHROW( projectItem->Checkin( NULL, NULL, 0 ) );
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        // clear case

        // get windows (and clearcase) username
        char userName[200];
        unsigned long userNameSize = 200;
        GetUserName( userName, &userNameSize );

        // get list of checked out files
        CComObjPtr<ICCCheckedOutFileQuery> checkedOutFileQuery;

        COMTHROW( m_clearCase->CreateCheckedOutFileQuery( PutOut(checkedOutFileQuery) ));

        VARIANT v;
        v.vt = VT_ARRAY | VT_BSTR;
        SAFEARRAYBOUND rgsabound[1];  //Denotes number of dimensions
        rgsabound[0].lLbound   = 0;
        rgsabound[0].cElements = 1;
        v.parray = SafeArrayCreate(VT_VARIANT, 1, rgsabound);
        long count = 0;

        CComVariant name = m_folderPath.c_str();
        COMTHROW( SafeArrayPutElement(v.parray, &count, (void *) &name) );
        
        // get list of checked out files
        CComBSTR user = userName;                
        COMTHROW( checkedOutFileQuery->put_PathArray( v ) );
        COMTHROW( checkedOutFileQuery->put_User(user) );
        CComObjPtr<ICCCheckedOutFiles> files;
        COMTHROW( checkedOutFileQuery->Apply( PutOut(files) ));

        // checkin files
        long fileCount;
        COMTHROW( files->get_Count( &fileCount ) );
        for( int i=0; i<fileCount; ++i )
        {
            CComObjPtr<ICCVersion>    version;
            CComBSTR                  comment = "";

            CComObjPtr<ICCCheckedOutFile> checkedOutFile;
            COMTHROW( files->get_Item( i+1, PutOut(checkedOutFile) ) );
            checkedOutFile->CheckIn( comment, 1, comment, ccKeep, PutOut(version) );
            
            // check out if keepcheck out is true
            if( keepCheckedOut )
            {            
                CComBSTR path;
                checkedOutFile->get_Path(&path);
                char buf[MAX_PATH];
                sprintf(buf, "%S", path);
                checkOutFileCC(buf);
            }
        }
    }
}

bool CCoreXmlFile::createClearCaseApp()
{
    try
    {
        if( m_clearCase.p == NULL )
        {
            COMTHROW( m_clearCase.CoCreateInstance( CLSID_Application ) );
            COMTHROW( m_clearTool.CoCreateInstance( CLSID_ClearTool ) );
        }
        return true;
    }
    catch(...)
    {
        AfxMessageBox( "Cannot connect to ClearCase. Check the ClearCase installation!" );
        return false;
    }
}

bool CCoreXmlFile::isPathUnderClearCase( const char * path )
{   
    ASSERT( m_clearCase.p != NULL );

    CComObjPtr<ICCVersion> version;

    CComBSTR path2 = path;
    VARIANT  path3;

    path3.bstrVal = path2.m_str;
    path3.vt      = VT_BSTR;

    HRESULT hres = m_clearCase->get_Version( path3, PutOut(version) );

    return (hres==S_OK);
}

bool CCoreXmlFile::isFileCheckedOutCC( const char * path )
{
    CComObjPtr<ICCVersion>        version;
    CComBSTR                      path1 = path;
    VARIANT                       path2;
    //HRESULT                       hres;
    VARIANT_BOOL                  isCheckdOut = 0;

    path2.vt      = VT_BSTR;
    path2.bstrVal = path1;
   
    COMTHROW( m_clearCase->get_Version( path2, PutOut(version) ) );
    COMTHROW( version->get_IsCheckedOut( &isCheckdOut ) );

    return (isCheckdOut!=0);
}

int CCoreXmlFile::getCheckOutStateCC( const char * path )
{
    // get windows (and clearcase) username
    char userName[200];
    unsigned long userNameSize = 200;
    GetUserName( userName, &userNameSize );

    // get list of checked out files by current user
    CComObjPtr<ICCCheckedOutFileQuery> checkedOutFileQuery;

    COMTHROW( m_clearCase->CreateCheckedOutFileQuery( PutOut(checkedOutFileQuery) ));

    VARIANT v;
    v.vt = VT_ARRAY | VT_BSTR;
    SAFEARRAYBOUND rgsabound[1];  //Denotes number of dimensions
    rgsabound[0].lLbound   = 0;
    rgsabound[0].cElements = 1;
    v.parray = SafeArrayCreate(VT_VARIANT, 1, rgsabound);
    long count = 0;

    CComVariant name = m_folderPath.c_str();
    COMTHROW( SafeArrayPutElement(v.parray, &count, (void *) &name) );
    
    CComBSTR user = userName;
    COMTHROW( checkedOutFileQuery->put_PathArray( v ) );
    COMTHROW( checkedOutFileQuery->put_User(user) );
    CComObjPtr<ICCCheckedOutFiles> files;
    COMTHROW( checkedOutFileQuery->Apply( PutOut(files) ));

    long fileCount;
    COMTHROW( files->get_Count( &fileCount ) );
    for( int i=0; i<fileCount; ++i )
    {
        CComObjPtr<ICCVersion>    version;
        CComBSTR                  comment = "";

        CComObjPtr<ICCCheckedOutFile> checkedOutFile;
        COMTHROW( files->get_Item( i+1, PutOut(checkedOutFile) ) );

        CComBSTR path2;
        checkedOutFile->get_Path(&path2);
        char path3[MAX_PATH];
        sprintf(path3, "%S", path2);

        if( stricmp(path, path3) == 0 )
        {
            return CS_CURRENT_USER;
        }
    }

    if( isFileCheckedOutCC( path ) )
        return CS_OTHER_USER;

    return CS_NOT_CHECKEDOUT;
}

void CCoreXmlFile::checkOutFileCC( const char * path )
{
    CComObjPtr<ICCVersion>        version;
    CComObjPtr<ICCCheckedOutFile> checkedOutFile;
    CComBSTR                      path1 = path;
    CComBSTR                      comment = "";
    VARIANT                       path2;
    //HRESULT                       hres;

    path2.vt      = VT_BSTR;
    path2.bstrVal = path1;
   
    COMTHROW( m_clearCase->get_Version( path2, PutOut(version) ) );
    version->CheckOut( ccTryReserved, comment, 0, ccVersion_Default, 0, 0, PutOut(checkedOutFile) );
    //COMTHROW( version->CheckOut( ccReserved, comment, 0, ccVersion_Default, 0, 0, PutOut(checkedOutFile) ) );
}

void CCoreXmlFile::checkInFileCC( const char * path )
{
    CComObjPtr<ICCVersion>        version;
    CComObjPtr<ICCCheckedOutFile> checkedOutFile;
    CComBSTR                      path1 = path;
    CComBSTR                      comment = "";
    CComBSTR                      empty = "";
   
    if( m_clearCase->get_CheckedOutFile( path1, PutOut(checkedOutFile) ) == S_OK )
    {
        if( checkedOutFile.p != NULL )
        {
            HRESULT hres = checkedOutFile->CheckIn( comment, 1, empty, ccKeep, PutOut(version) );
        }
    }
}

void CCoreXmlFile::addFileToCC( const char * path )
{
    // parent folder must be checked out!
    CComObjPtr<ICCCheckedOutFile> checkedOutFile;

    CComBSTR fileName         = path;
    CComBSTR comment          = "";
    CComBSTR elementTypeName1 = "text_file";

    VARIANT elmentTypeName;
    elmentTypeName.vt = VT_BSTR;        
    elmentTypeName.bstrVal = elementTypeName1;

    m_clearCase->CreateElement(fileName,comment,1,elmentTypeName,PutOut(checkedOutFile)); 
}

void CCoreXmlFile::addDirToCC( const char * path )
{
    // parent folder must be checked out!
    CComObjPtr<ICCCheckedOutFile> checkedOutFile;

    CComBSTR fileName         = path;
    CComBSTR comment          = "";
    CComBSTR elementTypeName1 = "directory";

    VARIANT elmentTypeName;
    elmentTypeName.vt = VT_BSTR;        
    elmentTypeName.bstrVal = elementTypeName1;

    m_clearCase->CreateElement(fileName, comment,1,elmentTypeName,PutOut(checkedOutFile));
}

void CCoreXmlFile::getLatestVerCC( const char * path )
{
    std::string cmd = "update ";
    cmd += path;
    CComBSTR cmd2 = cmd.c_str();
    BSTR output;
    COMTHROW( m_clearTool->CmdExec( cmd2, &output ) );
}

void CCoreXmlFile::showUsedFiles( XmlObjSet& containers )
{
    if( m_sourceControl == SC_SOURCESAFE )
    {
        CFilesInUseDetailsDlg dlg;
        char buf[300];
        XmlObjSet::iterator it;
        for( it=containers.begin(); it!=containers.end(); ++it )
        {
            string user;
            getSSCheckOutUser( *it, user );
            if( user.size() > 0 )
            {
                string name, type;
                getContainerName( *it, name, type );
                sprintf( buf, "%s\t%s (%s)", user.c_str(), name.c_str(), type.c_str() );
                dlg.m_fileList.push_back( buf );
            }
        }
        dlg.DoModal();
    }
    else 
    {
        AfxMessageBox( "No detailed information available." );
    }
}

void CCoreXmlFile::setSourceControlInfo( XmlObject * container, const char * scInfo )
{
    ASSERT( container->isContainer() );

    // find regnode and set it if found
    for(AttribMapIter it=container->m_attributes.begin(); it!=container->m_attributes.end(); ++it)
    {
        if( it->second->getType() == VALTYPE_COLLECTION )
        {
            XmlAttrCollection * coll = (XmlAttrCollection *)it->second;
            for( XmlObjSetIter it2=coll->m_children.begin(); it2!=coll->m_children.end(); ++it2 )
            {
                XmlObject * obj2 = (*it2);
                if( obj2->m_sourceControlInfoRegNode )
                {
                    AttribMapIter it = obj2->m_attributes.find( ATTRID_REGNODEVALUE );
                    ASSERT( it!=obj2->m_attributes.end() );
                    it->second->fromString(scInfo);
                    return;
                }
            }
        }
    }
}

void CCoreXmlFile::getSourceControlInfo( XmlObject * container, string& scInfo )
{
    ASSERT( container->isContainer() );

    scInfo = "";

    // find regnode and set it if found
    for(AttribMapIter it=container->m_attributes.begin(); it!=container->m_attributes.end(); ++it)
    {
        if( it->second->getType() == VALTYPE_COLLECTION )
        {
            XmlAttrCollection * coll = (XmlAttrCollection *)it->second;
            for( XmlObjSetIter it2=coll->m_children.begin(); it2!=coll->m_children.end(); ++it2 )
            {
                XmlObject * obj2 = (*it2);
                if( obj2->m_sourceControlInfoRegNode )
                {
                    AttribMapIter it = obj2->m_attributes.find( ATTRID_REGNODEVALUE );
                    ASSERT( it!=obj2->m_attributes.end() );
                    it->second->toString(scInfo);
                    return;
                }
            }
        }
    }
}

void CCoreXmlFile::updateSourceControlInfo( XmlObject * container )
{
    ASSERT( container->isContainer() );

    if( m_sourceControl == SC_SOURCESAFE )
    {
        string user;
        getSSCheckOutUser( container, user );
        if( user.size() == 0 )
            setSourceControlInfo( container, "GME_SourceControl_State_none" );  // not checked out
        else if( stricmp( user.c_str(), m_vssUser.c_str() ) == 0 )
            setSourceControlInfo( container, "GME_SourceControl_State_local" );  // checked out by local user
        else
            setSourceControlInfo( container, "GME_SourceControl_State_other" );  // checked out by other user
    }
    else if( m_sourceControl == SC_CLEARCASE )
    {
        string fileName;
        getContainerFileName( container, fileName, true );
        int state = getCheckOutStateCC( fileName.c_str() );
        if( state == CS_NOT_CHECKEDOUT )
            setSourceControlInfo( container, "GME_SourceControl_State_none" );
        else if( state == CS_CURRENT_USER )
            setSourceControlInfo( container, "GME_SourceControl_State_local" );
        else
            setSourceControlInfo( container, "GME_SourceControl_State_other" );
    }
}

void CCoreXmlFile::updateSourceControlInfo()
{
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);
        if( obj->isContainer() )
            updateSourceControlInfo( obj );
    }
    //dumpSourceControlInfo();

    //if( m_gme.p != NULL )
      //  m_gme->CloseProject( TRUE );
}

void CCoreXmlFile::dumpSourceControlInfo()
{
    FILE * f = fopen( "c:\\temp\\out.txt", "w" );

    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);
        if( obj->isContainer() )
        {
            string guid, str;
            guid2str( obj->m_guid, guid );
            getSourceControlInfo( obj, str );

            fprintf( f, "%s\t%s\n", guid.c_str(), str.c_str() );
        }
    }

    fclose(f);
}

bool CCoreXmlFile::filesModifedByOthers()
{
    for( XmlObjVecIter it=m_objects.begin(); it!=m_objects.end(); ++it )
    {   
        XmlObject * obj = (*it);
        if( obj->isContainer() )
        {
            string filename;
            getContainerFileName(obj, filename);

            // get last write time
            WIN32_FILE_ATTRIBUTE_DATA attr;        
            GetFileAttributesEx( filename.c_str(), GetFileExInfoStandard, &attr );
            CTime lastWriteTime( attr.ftLastWriteTime );

            if( lastWriteTime > obj->m_lastWriteTime )
                return true;           
        }
    }
    return false;
}


// egyeb:
// transaction-ban nem lehet ki-be csekkelni es get latestet venni!!!

// objektum toreles esten felulrol kilancoljak az objektumot es nem is torlik csak az egesz
// projekt bezarasanal -> szivas?


/*
TODO:
1. Az open-nel lehet hogy mgx-filet kene hasznalni.
   Kerdes hogyan fogjuk soruce control ala rakni a projektet.
   - lehetne a create-nel megadni es a connection stringben atmenhet az info
   - ha nem ott akkor kesobb kell egy eszkosz ami ezt megteszi, ez viszont a core-t kikerulve
     tudja csak az mgx file-ba beirni a szukseges infot


2. minden muvelet-hez (put_attrib,delete_object) meg kell tudni hatarozni a kicsekelendo file-ok halmazat
    - tobb algoritmus van
        - szigoru (osszes gyerek kicsekkelve)

   - ha transaction kezdodik elekzdjuk gyujteni az osszes modosult objektet

   esetek:
   1. sima attr modositas:          az objektumot megjegyezzuk
   2. nem parent pointer modositas: az objektumot megjegyezzuk
   3. parent pointer modositas: 
    - a regi containert megjegyezzuk
    - az uj containert megjegyezzuk

6. Mi van ha egy atomot atrakunk egy A modellbol egy B-be (megvaltoztatjuk a parentjet)
   ez kemeny szivas:
   - masik file-ba kerul, a regi containert is ki kell csekkelni es torolni belole
   - az uj kontainert is ki kell csekkelni

7. A ATTRID_LASTRELID miatt kicsekkelodik a szulo is ha beleszurok egy gyereket

8. Az MgaProject-ben az abortTransaction-ben a popTerritory nincs korrektul megcsinalva!

9. A GUID rendes attributum legyen

10. a ki-be csekkeltsegi allapot is attributum legyen hogy a browser tudja mutatni

11. Amikor tobb file-t egyszerre csekkelek ki transaction szeruen kellene
    - Ez jelenle nem olyan nagy problema, mert ellenorizzuk hogy ki tudjuk-e csekkelni mindet
      es csak akkor kerzdunk neki ha igen. Annak a valoszinusege hogy nem sikerul nagyon kicsi

12. Amikr kicsekkelem a dependencyt azt a regi graf szerint teszem, az uj szeretin kellene...
  - megoldas: minden kicsekk elott getlatest, es az uj fileokat processzalni kell

13. GUI-t ki kell venni a backend-bol, errorinfot kell atadni stb.

14. MGX connection string-be infok: crete-nel source control info, opennel username password

  Create:
    MGX="C:\temp\test1" vssdatabase="frfre" vsspath="dede" user="fdasfdsfds" password="defrefre"

19. ha idezojeleket hasznalunk a nevekben, az xml serializacio miatt nem mukszik a dolog jol

20. Autosave-nel minden kicsekkelve marad, sima save-nel meg kell kerdezni
    // a close-ban duplan van becsekkeles ezzel valamit csinalni kell

21. GMEView bugs: ha ki van csekkelve masnal akkor az abort transaction a kovetkezo esetkben nem 
    mukodik jol
    - ctrl+move
    - torles a fabol jobb gombal
    - ha a faba masolunk, a fat frissiteni kell utana

22. binary chache-el van valami bibi





TODO:
1. cache tesztelese nagyobb projekteknel
2. hibakalezels
     
 


SZIVAS!

  
  Kerdesek: 
  1. a lockot minek undozni?



ClearCase kerdesek:
- a directory-k mergelese: 
- dynamic view eseten nincs szukseg get latestre
- a parent konyvtarat egyszeruen kicsekkelve kell tartani az ownernek es kesz
  igy mindeki hozza tud adni fileokat (tesztek szerint igy mukodik)


Hibak:
-------
1. Ha hozzadunk egy uj modelt akkor lehet hogy olyan helyre kerul ahol vannak mas modellek
   amik ki vannak csekkelve masok altal amit az autoaranelment miatt majd megprobal kicsekkelni


TODO: kell hogy mutassuk a browserbe hogy mit lehet editalni
 1-es megodlas: bevezetunk egy virtualis attributumot, regnodot
  - le tudjuk-e ezt kerdezni felulrol???



 

*/