/*
 * Copyright (c) 2002, Vanderbilt University
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE VANDERBILT UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE VANDERBILT
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE VANDERBILT UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE VANDERBILT UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: Gyorgy Balogh
 * Date last modified: 10/23/03
 */

/*
 * TestMgaLayer1.java
 *
 * Created on September 17, 2003, 10:08 AM
 */

package org.isis.gme.mga;

import org.isis.gme.meta.MgaMetaAspect;
import org.isis.gme.meta.MgaMetaAspects;
import org.isis.gme.meta.MgaMetaAttribute;
import org.isis.gme.meta.MgaMetaFCO;
import org.isis.gme.meta.MgaMetaModel;
import org.isis.gme.meta.MgaMetaPart;
import org.isis.gme.mgautil.MgaRegistrar;
import org.isis.jaut.Apartment;
import org.isis.jaut.Dispatch;

/**
 *
 * @author  bogyom
 */
public class TestMgaLayer1
{
    private void printAspectNames( MgaModel model )
    {
        System.out.println( "aspect of " + model.getName() );
        MgaMetaModel metaModel = model.getMetaModel();
        MgaMetaAspects aspects = metaModel.getAspects();
        for( int i=0; i<aspects.getCount(); ++i )
            System.out.println( "  " + aspects.getItem(i).getName() );
    }
    
    private void showPorts( MgaModel model, MgaMetaAspect aspect )
    {      
        MgaPart parts[] = model.getAspectParts(aspect).getAll();
        for( int i=0; i<parts.length; ++i )
        {   
            MgaFCO      fco      = parts[i].getFCO();
            MgaMetaPart metaPart = new MgaMetaPart( parts[i].getMeta() );        
            if( metaPart.getIsLinked() )
            {                        
                System.out.println( "    " + fco.getName() );                                
            }                       
        }  
    }       
        
    private void showModel( MgaModel model, MgaMetaAspect aspect )
    {
        System.out.println( "\nmodel: " + model.getName() + " aspect: " + aspect.getName() );

        MgaPart parts[] = model.getAspectParts(aspect).getAll();
        for( int i=0; i<parts.length; ++i )
        {   
            MgaFCO           fco      = parts[i].getFCO();
            MgaMetaFCO       metaFco  = fco.getMeta();           
            int              type     = fco.getObjType();
            MgaPart.GmeAttrs attr     = parts[i].getGmeAttrs();
            MgaMetaPart      metaPart = new MgaMetaPart( parts[i].getMeta() );
            
            String iconPath = metaFco.getRegistryValue("icon");
                      
                       
            if( type == MgaObject.OBJTYPE_ATOM )
            {
                System.out.println( "atom: " + fco.getName() + "," + attr.x + "," + attr.y + "," + iconPath );                                               
            }
            else if( type == MgaObject.OBJTYPE_MODEL )
            {
                System.out.println( "model: " + fco.getName() + "," + attr.x + "," + attr.y + "," + iconPath );
                
                // find out aspect for childs
                MgaMetaModel metaModel = new MgaMetaModel( (new MgaModel( fco )).getMetaModel() );            
                String kindAspect = metaPart.getKindAspect();
                if( kindAspect.length() == 0 )
                    kindAspect = aspect.getName();                                         
                MgaMetaAspect aspectForChildren = metaModel.getAspectByName( kindAspect );
                                                        
                // show ports
                showPorts( new MgaModel(fco), aspectForChildren );                              
            }
            else if( type == MgaObject.OBJTYPE_REFERENCE )
            {
                //getReferred()                
            }           
            else if( type == MgaObject.OBJTYPE_CONNECTION )
            {                
                MgaSimpleConnection conn = new MgaSimpleConnection( fco );               
                System.out.println( "connection from " + conn.getSrc().getName() + " to " + conn.getDst().getName() );                                  
            }                       
        }
    }
    
    private void printName( MgaObject obj , String prefix )
    {
        System.out.println( prefix + obj.getName() );
        
        int type = obj.getObjType();
        if( type == 1 || type == 6 )
        {
            MgaObjects children = obj.getChildObjects();
            for( int i=0; i<children.getCount(); i++ )
                printName( children.getItem(i) , prefix + "  " );
        }
    }
    
    private String getIconFileName( MgaFCO fco )
    {              
        MgaMetaFCO metaFco = fco.getMeta();           
            
        String iconFile = metaFco.getRegistryValue("icon");        
        
        MgaRegistrar reg = new MgaRegistrar();
        String iconPath  = reg.getIconPath(MgaRegistrar.REGACCESS_BOTH);
        
        //getProjectConnStr()
        //getParadigmConnStr()
        
        MgaProject proj = fco.getProject();
        
        
        return "";
    }
    
    private void printAttributes( MgaFCO fco )
	{
		System.out.println( "Attributes of " + fco.getName() +":" );
		
		MgaAttribute [] attribs = fco.getAttributes().getAll();
		for( int i=0; i<attribs.length; ++i )
		{
            MgaMetaAttribute metaAttr = attribs[i].getMeta();
            
    		System.out.println( "    " + metaAttr.getName() + " = " +  attribs[i].getValue() );
        }
	}
    
    private void test1()
    {
        Apartment.enter(true);       
        
        // create and open mga project
        MgaProject project = MgaProject.createInstance();
        project.open( "MGA=C:\\Temp\\gmetest\\alma3.mga" );
        //project.open( "MGA=C:\\Temp\\gmetest\\SFDemo.mga" );
        //project.open( "MGA=C:\\Temp\\gmetest\\MetaGME-model.mga" );
        project.beginTransaction( null );
        
        // write object structure
        MgaObject root = new MgaObject( project.getRootFolder() );
        //printName( root, "" );               
            
        // show model        
        
        //MgaModel model = new MgaModel( root.getChildObjects().getItem(0).getChildObjects().getItem(14) );
        //MgaModel model = new MgaModel( root.getChildObjects().getItem(0));
        //MgaMetaModel metaModel = model.getMetaModel();
        //MgaMetaAspect aspect = metaModel.getAspects().getItem(0);        
        //printAspectNames( model );
        //showModel( model, aspect );
		//printAttributes( model );
        
        //System.out.println( project.getName() );
        
        //MgaFCO fco = new MgaModel( root.getChildObjects().getItem(0) );
        //MgaMetaFCO metaFCO = fco.getMeta();
        
        //MgaFolder folder = new MgaFolder( root );//.getChildObjects().getItem(0));
        //MgaFCO fco1 = folder.createRootObject( metaFCO );
        //fco1.setName("ez az uj jatekos");
        
        printName( root, "" );
        
              
        
        //System.out.println( fco1.getName() );
        
                                       
        
        //project.commitTransaction();
        project.abortTransaction();
        //project.save();
        project.close();
        Apartment.leave();
        System.out.println("done"); 
    }

    void test3()
    {
        MgaProject project = MgaProject.createInstance();            
        
//        project.open( "MGA=C:\\Temp\\gmetest\\alma3.mga" );
        
//        project.beginTransaction( null );
//        System.out.println(project.getName());
//        project.commitTransaction();
    
        project.close();
    }
    
    void test2()
    {
        try
        {
            Apartment.enter(false);
            test3();
            Apartment.leave();
            
            System.out.println("done");
        }
        catch(Exception e)
        {
            e.printStackTrace();
        }
    }
    
    private void xmlBackEndTest1()
    {
        Apartment.enter(true);       
        
        MgaProject project = MgaProject.createInstance();
        project.open( "MGA=C:\\Temp\\gmetest\\alma2.mga" );
        
        //project.beginTransaction( null );        
        project.save( "MGA=C:\\Temp\\gmetest\\alma123.mga", false );
        project.save( "XML=C:\\Temp\\gmetest\\alma124.mga", false );
        
        //project.commitTransaction();
       
        project.close();
        Apartment.leave();
        System.out.println("done");         
    }
    
    public static void bonTest( int project, int currentObj, int selectedObjs, int param )
    {
        try
        {
            Apartment.enter(false);
                   
            MgaProject p = new MgaProject( new Dispatch(project) );
            
            System.out.println( "A project neve:" + p.getName() );            
            
            Apartment.leave();
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }
    }
    
    public static void main(String[] args)
    {
        TestMgaLayer1 t = new TestMgaLayer1();
        t.test2();
        //t.xmlBackEndTest1();
    }
    
}
